import numpy as np
from copy import deepcopy

from mpl_plotter import figure
from mpl_plotter.two_d import line, scatter, comparison
from mpl_plotter.color.schemes import colorscheme_one
from mpl_plotter.color.functions import delta

from alexandria.shell import print_color, print_result
from alexandria.data_structs.string import join_set_distance

from huracan.constants import R


class component:
    """
    Component
    ---------
    """
    def __sub__(self, other):
        """
        Stream creation operator: <component> - <component>
        """
        if isinstance(other, component):
            s = stream()-other
            return s

    def __call__(self, gas):
        """
        Component transfer function execution
        """
        p = self.tf(gas)
        for k, v in p.__dict__.items():
            if k[-2:] == '01':
                k = k[0] + '0'
            setattr(self, k, v)

        # Gas state variables
        for sv in ['V', 'S', 'H']:
            setattr(self, sv, getattr(gas, sv))


class shaft:
    """
    Shaft
    -----
    """
    def __init__(self, *args, eta, eta_gearbox=1):
        """
        :param args: list of components connected by the shaft.
        :param eta:  mechanical efficiency of the shaft.

        :type args:  component
        :type eta:   float
        """
        self.eta         = eta
        self.eta_gearbox = eta_gearbox
        self.components  = list(args)

        for c in args:
            c.shaft = self

    def w_exerting_machinery(self):
        """
        Return a list of all components in the shaft
        which exert work on the flow. That is, instances
        of the fan and compressor classes.
        """
        return [c for c in self.components if c.__class__.__name__ in ['fan',
                                                                       'prop',
                                                                       'propfan',
                                                                       'compressor']]

    def electrical_plants(self):
        return [c for c in self.components if c.__class__.__name__ in ['power_plant']]

    def w_r(self):
        """
        Obtain the work required by the components which
        exert work on the gas (fan, compressors).
        """
        wem = self.w_exerting_machinery()

        assert all([hasattr(c, 'w') for c in wem]), \
            "The shaft's work exerting components do not have " \
            "a work attribute: ensure the streams to which each " \
            "belongs have been run up to the respective work " \
            "exerting component."

        work = np.array([c.w/self.eta_gearbox if c.__class__.__name__ in ['fan', 'prop', 'propfan']
                         else c.w for c in wem])
        etas = np.array([c.shaft.eta for c in wem])
        w_r_m = np.sum(work/etas)                           # Power required by work exerting components

        electrical = self.electrical_plants()               # FIXME: ugly
        w_r_e = sum([c.w_r for c in electrical])            # Power required by all electrical plants

        return w_r_m + w_r_e


class stream:
    """
    Stream
    ------
    """
    def __init__(self,
                 gas=None,
                 parents=None,
                 fr=None):
        """
        :param fr:      Fraction of the gas instance passed
                        to the stream which physically enters
                        the stream.
                        This is useful so the original gas
                        instance can be passed to child streams
                        in a stream diversion process.
                        In this way, the gas attribute of the
                        child streams points to the original
                        stream's gas instance until the moment
                        the child streams are run: at this
                        time, a deep copy of the original gas
                        instance is created, and the mass flow
                        multiplied by _fr_ to reflect the mass
                        flow actually flowing in the child stream.
        :param parents: Parent streams.
                        - If parents includes 2 or more streams,
                          they will be merged at runtime.

        :type gas:     gas
        :type fr:      float
        :type parents: list of stream
        """
        self.stream_id  = [0]
        self.components = []
        self.downstream = [self]

        self.ran = False

        if not isinstance(gas, type(None)):
            self.gas    = gas
        if not isinstance(parents, type(None)):
            self.parents = parents

        # Runtime dictionary
        self.runtime = {}
        if not isinstance(fr, type(None)):
            self.runtime['fr'] = fr

    """
    Operators
    """
    def __call__(self, gas):
        self.gas = gas
        return self

    def __sub__(self, other):
        """
        Stream concatenation operator: <stream> - <component/stream>
        """
        if isinstance(other, component):
            self._add_component(other)
            return self
        if isinstance(other, stream):               # TODO: stream merge
            return self._add_stream(other)

    def _add_component(self, c):
        """
        Component addition
        """
        self.components.append(c)
        c.downstream = self.downstream
        c.stream = self

    def _add_stream(self, s):
        """
        Stream addition
        """
        assert hasattr(self, 'gas') and hasattr(s, 'gas'), 'Both streams must have a gas attribute for' \
                                                           'the stream merge operation to be possible.'

        n = max(self.stream_id[0], s.stream_id[0])  # Get largest stream_id
        self.stream_id[0] = s.stream_id[0] = n      # Set largest stream_id for both merging streams

        merged = stream(parents=[self, s])
        merged.stream_id[0] = n + 1

        if hasattr(self, 'system') and hasattr(s, 'system'):
            self.system = s.system = merged.system = self.system + s.system
            self.system(merged)
        elif hasattr(self, 'system'):
            self.system(s, merged)
        elif hasattr(s, 'system'):
            s.system(self, merged)
        else:
            system(self, s, merged)

        return merged

    def __mul__(self, other):                       # TODO: stream diversion
        """
        Stream diversion operator: <stream> * n     for n: 0 =< float =< 1
        """
        return self._divert(other)

    def _divert(self, fr, names=None):
        """
        Stream diversion
        """

        assert hasattr(self, 'gas'), 'The stream must have a gas attribute for' \
                                     'the stream diversion operation to be possible.'

        main = stream(self.gas, fr=fr, parents=[self])
        div  = stream(self.gas, fr=1-fr, parents=[self])

        # Stream ID
        main.stream_id[0] = self.stream_id[0] + 1
        div.stream_id[0]  = self.stream_id[0] + 1

        # Diverted stream IDs
        if not isinstance(names, type(None)):
            main.stream_id.append(names[0])
            div.stream_id.append(names[1])
        else:
            mf_matrix = np.array([[main.gas.mf * fr,     main],
                                  [div.gas.mf  * (1-fr), div]])
            mf_matrix = mf_matrix[mf_matrix[:, 0].argsort()]

            for i in range(mf_matrix[:, 1].size):
                sub_id = 'm' if i == 0 else f's{i}' if i > 1 else 's'
                mf_matrix[i, 1].stream_id.append(sub_id)

        if hasattr(self, 'system'):
            self.system(main, div)
            main.system = div.system = self.system
        else:
            system(self, main, div)

        return main, div

    def __getitem__(self, item):
        """
        Component retrieval operator: <stream>[<component stage>]

        :type item: str
        """
        return self._system_takeover('__getitem__', item)

    def ___getitem__(self, item):
        """
        Retrieve any stream component by its stage name.

        :type item: str
        """
        assert item in self.stages(), 'Specified a non-existent stage.'

        for c in self.components:
            if c.stage == item:
                return c

    """
    Utilities
    """
    def stages(self):
        """
        Return a list containing the stage name of each
        component in the stream.
        """
        return [c.stage for c in self.components]

    def _stage_name(self, c):
        """
        Return the stage name of a component in the stream,
        composed of the stream identification number, a code
        representing its parent class, and a numerical index
        if there are more than 1 components of the same class
        in the stream.
        """
        codes = {'fan':                'fn',
                 'prop':               'pr',
                 'propfan':            'pf',
                 'intake':             'it',
                 'inlet':              'il',
                 'compressor':         'cp',
                 'combustion_chamber': 'cc',
                 'turbine':            'tb',
                 'afterburner':        'ab',
                 'nozzle':             'nz'
                 }

        code = codes[c.__class__.__name__] + self._n_instances(c)

        return f'{".".join([str(c) for c in self.stream_id])}.{code}'

    def _n_instances(self, comp):
        """
        Calculate the number of instances of a given component's
        parent class in the stream (n), and its index in the
        stream's components list (i).

        The index of the component is returned as follows:
        - If the given component is the only instance of its parent
          class in the stream (n = 1):
            - ''                                     (empty string)
        - If the given component is one of more instances of its
          parent class in the stream (n > 1):
            - str(i + 1)                    (numeral starting at 1)

        :type comp: component
        """

        i = 0       # Component index
        n = 0       # Number of instances of the input component's class in the stream
        for c in self.components:
            if comp is c:
                i = n
            if comp.__class__.__name__ == c.__class__.__name__:
                n += 1

        return '' if n == 1 else str(i + 1)

    def _log(self):

        d = 9

        for c in self.components:
            section_name = join_set_distance(c.stage, c.__class__.__name__.capitalize().replace("_", " "), d)
            print_color(section_name, 'green')

            if c.__class__.__name__ == 'nozzle':
                if c.choked:
                    print_color(' '*d + 'Choked flow', 'red')
            print_result(' '*(d+1) + 'T0', c.t0, '[K]')
            print_result(' '*(d+1) + 'p0', c.p0, '[Pa]')

    """
    Stream runtime functions
    """
    def _run(self, log=True):
        """
        Execute the transfer functions of all components in the stream
        on the instance's gas class instance.
        """

        self._runtime()

        assert hasattr(self, 'gas'), 'stream does not have a gas attribute.'

        self.choked = False                                 # FIXME: choked flow implementation is ugly

        for c in self.components:
            c(self.gas)                     # Run thermodynamic process on stream gas
            c.stage = self._stage_name(c)   # Set component stage name

            if hasattr(c, 'choked') and c.choked:           # FIXME: ugly
                self.choked = c.choked

        # Indicate stream has been run.
        self.ran = True

        if log:
            self._log()

    def _runtime(self):
        if hasattr(self, 'parents') and len(self.parents) > 1:
            self._merge()

        for k, v in self.runtime.items():
            f = getattr(self, '_'+k)
            f(v)

    def _merge(self):
        if hasattr(self, 'gas'):
            for s in self.parents:
                self.gas += s.gas
        else:
            self.gas = self.parents[0].gas
            for s in self.parents[1:]:
                self.gas += s.gas

    def _fr(self, fr):
        self.gas, _ = fr * deepcopy(self.gas)

    """
    Stream fluid state
    """
    def t0(self):
        """
        Total temperature vector.
        """
        assert self.ran, 'The stream must be run to obtain the total temperature at each stage'

        return np.array([c.t0 for c in self.components])

    def p0(self):
        """
        Total pressure vector.
        """
        assert self.ran, 'The stream must be run to obtain the total pressure at each stage'

        return np.array([c.p0 for c in self.components])

    def V(self):
        """
        Specific volume vector.
        """
        assert self.ran, 'The stream must be run to obtain the specific volume at each stage'

        return np.array([c.V for c in self.components])

    def S(self):
        """
        Specific entropy vector.
        """
        assert self.ran, 'The stream must be run to obtain the specific entropy at each stage'

        return np.array([c.S for c in self.components])

    """
    Stream outlet flow characteristics
    """
    def _v_exit(self):
        """
        Flow exit velocity

        Assumptions:
        - If the flow is not choked:
             The thermal energy lost by the gas as it leaves the nozzle
             is transformed into kinetic energy without losses.
        - If the flow is choked:
             The exit velocity is the velocity of sound before the nozzle
             exit.
        """
        # Absolute temperature before the stream exit (likely but not necessarily a nozzle)
        if len(self.components) > 1:
            # If the stream has more components than 1, the absolute temperature
            # after the component previous to the last one is taken.
            if self.components[-1].__class__.__name__ == 'nozzle':
                t_before_exit = self.components[-2].t0
            else:
                t_before_exit = self.components[-1].t0
        else:
            if hasattr(self, 'parents'):
                # If the stream has a single component and a parent stream or streams
                if len(self.parents) > 1:
                    # If the stream has more than a single parent stream, the gases
                    # of each parent are copied, merged and the absolute temperature
                    # of the resulting gas mixture is taken.
                    for i in range(len(self.parents)):
                        if i == 0:
                            g = deepcopy(self.parents[i].gas)
                        else:
                            g += deepcopy(self.parents[i].gas)
                    t_before_exit = g.t0
                else:
                    # If the stream has a single parent, the absolute temperature
                    # of the parent's gas is taken.
                    t_before_exit = self.parents[0].gas.t0
            else:
                # Is the stream has a single component and no parent streams,
                # it is assumed that the setup consists of a intake-nozzle
                # setup, and the absolute temperature of the moving gas is
                # taken.
                t_before_exit = deepcopy(self.gas).absolute().t01

        if self.choked:
            return (self.gas.k(t_before_exit)*R*t_before_exit)**0.5         # M=1 immediately before nozzle exit
        else:
            assert t_before_exit - self.gas.t0 > 0, 'The total temperature of the flow is lower before ' \
                                                    'the nozzle tha outside the engine: this happens due to the ' \
                                                    'compressors not providing enough energy to the flow. You must ' \
                                                    'either increase the pressure ratio of the compressors or ' \
                                                    'decrease the power extracted from the flow to solve the ' \
                                                    'inconsistency.'
            return (2*self.gas.cp(t_before_exit)*(t_before_exit - self.gas.t0))**0.5    # Heat -> Kinetic energy

    def _A_exit(self):
        """
        Nozzle exit area
        """
        return self.gas.mf*R*self.gas.t0/(self.gas.p0*self._v_exit())

    """
    Stream performance analysis
    """
    def _fmf(self):
        """
        Stream fuel mass flow
        """
        fmf = 0
        for c in self.components:
            if hasattr(c, 'fuel') and hasattr(c.fuel, 'mf'):
                fmf += c.fuel.mf
        return fmf

    def _flow_thrust(self):
        """
        Flow thrust

        If the flow is choked, the expansion of the gas contributes to the thrust of the flow.
        """
        if self.choked:
            return self.gas.mf * (self._v_exit() - self.gas.v_0) + self._A_exit() * (
                        self.gas.p0 - self.gas.p_0)
        else:
            return self.gas.mf * (self._v_exit() - self.gas.v_0)

    def _prop_thrust(self):
        """
        Propeller/propfan thrust
        """
        if any([c.__class__.__name__ in ['prop', 'propfan'] for c in self.components]):
            propellers = [c for c in self.components if c.__class__.__name__ in ['prop', 'propfan']]
            prop_thrust = sum([prop.thrust(self.gas.v_0) for prop in propellers])
        else:
            prop_thrust = 0
        return prop_thrust

    def _total_thrust(self):
        """
        Flow thrust plus propeller/propfan thrust
        """
        return self._flow_thrust() + self._prop_thrust()

    def _sfc(self):
        """
        Specific fuel consumption
        """
        return self.fmf()/self.flow_thrust()

    def _Q_in(self):               #TODO: verify and implement efficiency calculations
        """
        Heat provided to the flow.
        """
        q_provided = 0
        for c in self.components:
            if c.__class__.__name__ == 'combustion_chamber':
                q_provided += c.Q
        return q_provided

    def _W_req(self):
        """
        Work required from the flow.
        """
        w_required = 0
        for c in self.components:
            if c.__class__.__name__ in ['fan',
                                        'prop',
                                        'propfan',
                                        'compressor']:
                w_required += c.w
        return w_required

    def _E_balance(self):
        """
        Flow energy balance.
        """
        return self._Q_in() - self._W_req()

    def _E_prop(self):
        """
        Energy added to the flow for propulsion.
        """
        return self._v_exit()**2/2

    def _efficiency(self):
        """
        Engine efficiency
        """
        return self._E_balance()/self._Q_in()

    def _prop_efficiency(self):
        """
        Stream propulsive efficiency.
        """
        return self._E_prop()/self._Q_in()

    """
    Plots
    """
    def _plot_T_p(self,
                  show=False,
                  plot_label=None,
                  color=colorscheme_one()[0],
                  **kwargs):
        """
        Plot
        - Total pressure
        - Total temperature
        """

        figure((9, 5))

        defaults = {'x_label': 'p$_0$ [kPa]',
                    'y_label': 'T$_0$ [K]'}

        further_custom = {**defaults, **kwargs}

        self.plot_cycle_graph(self.p0()/1000, self.t0(),
                              color=color,
                              plot_label=plot_label,
                              show=show,
                              # Further customization
                              x_tick_ndecimals=2,
                              **further_custom)

    def _plot_p_V(self,
                  show=False,
                  plot_label=None,
                  color=colorscheme_one()[0],
                  **kwargs):
        """
        Plot
        - Total specific volume
        - Total pressure
        """

        figure((9, 5))

        defaults = {'x_label': 'v$_0$ [m$^3$/n]',
                    'y_label': 'p$_0$ [kPa]'}

        further_custom = {**defaults, **kwargs}

        self.plot_cycle_graph(self.V(), self.p0()/1000,
                              color=color,
                              plot_label=plot_label,
                              show=show,
                              # Further customization
                              y_tick_ndecimals=2,
                              **further_custom)

    def _plot_T_S(self,
                  show=False,
                  plot_label=None,
                  color=colorscheme_one()[0],
                  **kwargs):
        """
        Plot
        - Specific entropy
        - Total temperature
        """

        figure((9, 5))

        defaults = {'x_label': 'S [J/K/n]',
                    'y_label': 'T$_0$ [K]',}

        further_custom = {**defaults, **kwargs}

        self.plot_cycle_graph(self.S(), self.t0() / 1000,
                              color=color,
                              plot_label=plot_label,
                              show=show,
                              # Further customization
                              y_tick_ndecimals=2,
                              **further_custom)

    def plot_cycle_graph(self,
                         x, y,
                         plot_label,
                         x_label, y_label,
                         color=colorscheme_one()[0],
                         show=False,
                         **kwargs
                         ):
        """
        General plot composed of an MPL Plotter line and scatter plot.

        The default arguments plus any valid MPL Plotter line plotting
        class arguments can be passed to this function.
        """
        fig = kwargs.pop('fig', None)

        defaults = {
            # Specifics
            'point_size': 30,
            # Markers
            'marker': 'x',
            # Color
            'color': delta(color, -0.3),
            # Arrangement
            'zorder': 2,
            # Further customization
            'aspect': 1/2,
            'x_tick_number': 10,
            'y_tick_number': 10,
            'demo_pad_plot': True,
            'y_label_pad': 5,
        }

        further_custom = {**defaults, **kwargs}

        # Connecting lines
        line(   x=x, y=y,
                # Figure
                fig=fig,
                # Specifics
                line_width=1,
                # Color
                color=color, alpha=0.65,
                # Arrangement
                zorder=1)
        # Stages
        scatter(x=x, y=y,
                # Figure
                fig=fig,
                # Further customization
                plot_label=plot_label,
                x_label=x_label,
                y_label=y_label,
                show=show,
                **further_custom)

    """
    System takeover and API functions
    ---------------------------------
    """
    def _system_takeover(self, method, *args, **kwargs):
        if hasattr(self, 'system'):
            return getattr(self.system, method)(*args, **kwargs)
        else:
            return getattr(self, '_' + method)(*args, **kwargs)

    def run(self, log=True):
        args = locals()
        del args['self']

        self._system_takeover('run', **args)

    """
    Stream outlet flow characteristics
    """
    def v_exit(self):
        """
        Flow exit velocity

        Assumptions:
        - If the flow is not choked:
             The thermal energy lost by the gas as it leaves the nozzle
             is transformed into kinetic energy without losses.
        - If the flow is choked:
             The exit velocity is the velocity of sound before the nozzle
             exit.
        """
        return self._system_takeover('v_exit')

    def A_exit(self):
        """
        Nozzle exit area
        """
        return self._system_takeover('A_exit')

    """
    Stream performance analysis
    """
    def fmf(self):
        """
        Stream fuel mass flow
        """
        return self._system_takeover('fmf')

    def flow_thrust(self):
        """
        Flow thrust
        """
        return self._system_takeover('flow_thrust')

    def prop_thrust(self):
        return self._system_takeover('prop_thrust')

    def total_thrust(self):
        return self._system_takeover('total_thrust')

    def sfc(self):
        """
        Specific fuel consumption
        """
        return self._system_takeover('sfc')

    def Q_in(self):  # TODO: verify and implement efficiency calculations
        """
        Heat provided to the flow.
        """
        return self._system_takeover('Q_in')

    def W_req(self):
        """
        Work required from the flow.
        """
        return self._system_takeover('W_req')

    def E_balance(self):
        """
        Flow energy balance.
        """
        return self._system_takeover('E_balance')

    def E_prop(self):
        """
        Energy added to the flow for propulsion.
        """
        return self._system_takeover('E_prop')

    def efficiency(self):
        """
        Engine efficiency
        """
        return self._system_takeover('efficiency')

    def prop_efficiency(self):
        """
        Stream propulsive efficiency.
        """
        return self._system_takeover('prop_efficiency')

    """
    Plots
    """
    def plot_T_p(self,
                 show=False,
                 plot_label=None,
                 color=colorscheme_one()[0],
                 **kwargs):
        args = locals()
        args.pop('self', None)
        args.pop('kwargs', None)

        self._system_takeover('plot_T_p', **{**args, **kwargs})

    def plot_p_V(self,
                 show=False,
                 plot_label=None,
                 color=colorscheme_one()[0],
                 **kwargs):
        args = locals()
        args.pop('self', None)
        args.pop('kwargs', None)

        self._system_takeover('plot_p_V', **{**args, **kwargs})

    def plot_T_S(self,
                 show=False,
                 plot_label=None,
                 color=colorscheme_one()[0],
                 **kwargs):
        args = locals()
        args.pop('self', None)
        args.pop('kwargs', None)

        self._system_takeover('plot_T_S', **{**args, **kwargs})


class system:
    """
    System
    ------
    """
    def __init__(self, *args):
        """
        Create a system from two objects.

        :type args: stream
        """
        self.streams = []
        self._gobble(list(args))

    """
    Operators
    """
    def __call__(self, *args):
        """
        System stream addition operator: <system>(<list of streams>)

        :type args: stream
        """
        self._gobble(list(args))

    def _gobble(self, streams):
        """
        :type streams: list of stream
        """
        for s in streams:
            self.streams.append(s)
            s.system = self

    def __add__(self, other):
        """
        System addition operator: <system> + <stream/system>

        :type other: system
        """
        streams = list(set(self.streams) & set(other.streams))
        return system(*streams)

    def __getitem__(self, item):
        """
        Retrieve any stream component by its stage name.

        :type item: str
        """
        components = []
        for s in self.streams:
            components += s.components

        assert item in [c.stage for c in components], 'Specified a non-existent stage.'

        for c in components:
            if c.stage == item:
                return c

    """
    System functions
    """
    def run(self, log=True):
        """
        Run stream system.
        """

        self._sort_streams()

        n = 0
        while not all([s.ran for s in self.streams]):
            for s in self.streams:
                if s.stream_id[0] == n:
                    s._run(log)
            n += 1

    def _sort_streams(self):
        """
        Sort system streams based on their stream ID
        """
        ids     = [''.join(str(c) for c in stream.stream_id) for stream in self.streams]
        indexes = [float(id.replace('m', '.1').replace('s', '.2')) for id in ids]
        self.streams = [s for _, s in sorted(zip(indexes, self.streams))]

    def parents(self):
        """
        Return all system streams with children.
        Useful to not calculate thrust, exit velocity
        and other stream outlet values for streams
        flowing to children streams.
        """
        parents = []
        for s in self.streams:
            parents += s.parents if hasattr(s, 'parents') else []
        return parents

    """
    Stream outlet flow characteristics
    """
    def v_exit(self):
        """
        Flow exit velocity

        Assumptions:
        - If the flow is not choked:
             The thermal energy lost by the gas as it leaves the nozzle
             is transformed into kinetic energy without losses.
        - If the flow is choked:
             The exit velocity is the velocity of sound before the nozzle
             exit.
        """
        v = [s._v_exit() for s in self.streams if s not in self.parents()]
        return v[0] if len(v) == 1 else v

    def A_exit(self):
        """
        Nozzle exit area
        """
        a = [s._A_exit() for s in self.streams if 'nozzle' in [c.__class__.__name__ for c in s.components]]
        return a[0] if len(a) == 1 else a

    """
    System performance analysis
    """
    def fmf(self):
        return sum([s._fmf() for s in self.streams])

    def flow_thrust(self):
        return sum([s._flow_thrust() for s in self.streams if s not in self.parents()])

    def prop_thrust(self):
        return sum([s._prop_thrust() for s in self.streams])

    def total_thrust(self):
        return self.flow_thrust() + self.prop_thrust()

    def sfc(self):
        return self.fmf()/self.flow_thrust()

    def Q_in(self):  # TODO: verify and implement efficiency calculations
        """
        Heat provided to the flow.
        """
        return sum([s._Q_in() for s in self.streams])

    def W_req(self):
        """
        Work required from the flow.
        """
        return sum([s._W_req() for s in self.streams])

    def E_balance(self):
        """
        Flow energy balance.
        """
        return self.Q_in() - self.W_req()

    def E_prop(self):
        """
        Energy added to the flow for propulsion.
        """
        return sum([s._E_prop() for s in self.streams])

    def efficiency(self):
        """
        System efficiency
        """
        return self.E_balance()/self.Q_in()

    def prop_efficiency(self):
        """
        Stream propulsive efficiency.
        """
        return self.E_prop()/self.Q_in()

    """
    Plots
    """
    def _plot(self,
              x, y,
              x_scale=None, y_scale=None,
              x_label=None, y_label=None,
              show=False,
              plot_label=None,                  # When called from a _system_takeover the plot_label and color
              color=colorscheme_one()[0],       # arguments are passed to the function, but disregarded.
              **kwargs):
        """
        System plot
        -----------

        x and y are the stream parameters to be
        plotted for each stream.

        Process
        1. Create figure
        2. Create state variable and plotters vectors
            2.1 Parent connectors
        3. comparison call

        comparison call
            - fig=None, ax=None -> plot_cycle_graph -> fig in **kwargs keys
                - fig=None, ax=None -> line, scatter
                    - line, scatter plot onto active figure, axis

        :param x_scale: Scaling factor.
        :param y_scale: Scaling factor.

        :type x:        str
        :type y:        str
        :type x_scale:  float
        :type y_scale:  float
        """
        plotters   = []
        x_system   = []
        y_system   = []

        scales     = {'t0': 1,
                      'p0': 1/1000,
                      'V':  1,
                      'S':  1,
                      'H':  1}
        x_scale    = scales[x] if isinstance(x_scale, type(None)) else x_scale
        y_scale    = scales[y] if isinstance(y_scale, type(None)) else y_scale

        # 1. Create figure
        figure((9, 5))

        # 2. Create state variable and plotters vectors
        for stream in self.streams:

            # Plot defaults
            defaults = {'plot_label': f'{".".join([str(c) for c in stream.stream_id])}',
                        'x_label': x_label,
                        'y_label': y_label,
                        'color': colorscheme_one()[self.streams.index(stream)],
                        'zorder': 10-self.streams.index(stream)
                        }

            def gen_plotter(**defaults):
                """
                Returns a plotter using the defaults.
                Any keyword arguemnts passed to the
                _plot function overwrite the defaults.
                """
                return lambda x, y, **kwargs: stream.plot_cycle_graph(x=x, y=y, **{**kwargs, **defaults})

            x_stream = getattr(stream, x)()*x_scale
            y_stream = getattr(stream, y)()*y_scale

            plotters.append(gen_plotter(**defaults))
            x_system.append(x_stream)
            y_system.append(y_stream)

            # 2.1 Parent connectors
            if hasattr(stream, 'parents'):
                for parent in stream.parents:
                    x_parent = getattr(parent, x)()*x_scale
                    y_parent = getattr(parent, y)()*y_scale
                    # If the parent stream has no stages, get parent stream's gas state
                    p_x = x_parent[-1] if len(x_parent) != 0 else getattr(parent.gas, x)*x_scale
                    p_y = y_parent[-1] if len(y_parent) != 0 else getattr(parent.gas, y)*y_scale
                    if len(stream.components) > 0:
                        x_system.append(np.array([p_x, x_stream[0]]))
                        y_system.append(np.array([p_y, y_stream[0]]))
                        plotters.append(gen_plotter(color=defaults['color'], x_label=None, y_label=None))

        # 2.2 Remove streams with no stages
        mask_x = np.array([a.size != 0 for a in x_system])
        mask_y = np.array([a.size != 0 for a in y_system])
        mask = mask_x * mask_y  # Ensure that any x-y array pairs with an empty array are removed

        x_system = np.array(x_system, dtype='object')[mask].tolist()
        y_system = np.array(y_system, dtype='object')[mask].tolist()
        plotters = np.array(plotters, dtype='object')[mask].tolist()

        # 4. comparison call
        #     - fig=None, ax=None -> plot_cycle_graph -> fig in **kwargs keys
        #         - fig=None, ax=None -> line, scatter
        #             - line, scatter plot onto active figure, axis
        comparison(x=x_system, y=y_system, f=plotters,
                   legend_loc=(0.875, 0.425),
                   show=show,
                   **kwargs)

    def plot_T_p(self,
                 show=False,
                 plot_label=None,
                 color=colorscheme_one()[0],
                 **kwargs
                 ):
        args = locals()
        args.pop('self', None)
        args.pop('kwargs', None)

        self._plot(x='p0', x_label='p$_0$ [kPa]',
                   y='t0', y_label='T$_0$ [K]',
                   **{**args, **kwargs}
                   )

    def plot_p_V(self,
                 show=False,
                 plot_label=None,
                 color=colorscheme_one()[0],
                 **kwargs):
        args = locals()
        args.pop('self', None)
        args.pop('kwargs', None)

        self._plot(x='V',  x_label='v$_0$ [m$^3$/n]',
                   y='p0', y_label='p$_0$ [kPa]',
                   **{**args, **kwargs})

    def plot_T_S(self,
                 show=False,
                 plot_label=None,
                 color=colorscheme_one()[0],
                 **kwargs):
        args = locals()
        args.pop('self', None)
        args.pop('kwargs', None)

        self._plot(x='S',  x_label='S [J/K]',
                   y='t0', y_label='T$_0$ [K]',
                   **{**args, **kwargs})



from huracan.engine import component


class compressor(component):
    """
    Compressor
    ----------

    Adiabatic compression.
    """
    def __init__(self,
                 eta,
                 PI=None,
                 TAU=None):
        """
        :type eta: float
        :type PI:  float
        :type TAU: float
        """
        self.eta = eta
        self.PI  = PI
        self.TAU = TAU

    def tf(self, gas):
        return gas.compression(eta=self.eta, PI=self.PI, TAU=self.TAU)


class turbine(component):
    """
    Turbine
    -------

    Adiabatic expansion.
    """
    def __init__(self,
                 eta,
                 PI=None,
                 TAU=None):
        """
        :type eta: float
        :type PI:  float
        :type TAU: float
        """
        self.eta = eta
        self.PI  = PI
        self.TAU = TAU

    def tf(self, gas):
        """
        Before runtime, if no total pressure or temperature
        ratio has been provided, a minimum total temperature
        ratio is calculated.
        This temperature ratio is that required to power the
        work exerting components in the turbine's shaft.
        """
        if isinstance(self.PI, type(None)) and isinstance(self.TAU, type(None)):
            t00 = self.stream.gas.t0
            t01 = self.stream.gas.t0 - self.shaft.w_r()/(self.stream.gas.mf*self.stream.gas.cp(t00))
            self.TAU = t01/t00

        return gas.expansion(eta=self.eta, PI=self.PI, TAU=self.TAU)

    def w_r(self):
        return self.shaft.w_r()


class screw(component):
    """
    Screw
    -----

    Mechanical device to turn rotational to axial motion.
    """
    def __init__(self,
                 eta,
                 PI=None,
                 TAU=None):
        """
        :type eta: float
        :type PI:  float
        :type TAU: float
        """
        self.eta = eta
        self.PI  = PI
        self.TAU = TAU


class fan(screw):
    """
    Fan
    ---

    Fundamental differences with a propeller:
    - Disk loading.
    - Fan shroud.
    - Bypass flow.
    """

    def __init__(self,
                 eta,
                 PI=None,
                 TAU=None):
        """
        :type eta: float
        :type PI:  float
        :type TAU: float
        """
        super().__init__(eta=eta,
                         PI= PI,
                         TAU=TAU)

    def tf(self, gas):
        return gas.compression(eta=self.eta, PI=self.PI, TAU=self.TAU)


class prop(screw):
    """
    Propeller
    ---------
    """
    def __init__(self,
                 eta,
                 w,
                 PI=None,
                 TAU=None):
        """
        :param w:  [W] Propeller break power

        :type eta: float
        :type w:   float
        :type PI:  float
        :type TAU: float
        """
        super().__init__(eta=1,
                         PI=PI,
                         TAU=TAU)
        self.w        = w
        self.eta_prop = eta

    def thrust(self, v0):
        return self.w*self.eta_prop/v0

    def tf(self, gas):
        p = gas.compression(eta=self.eta, PI=self.PI, TAU=self.TAU)   # Placeholder
        p.w = self.w
        return p


class propfan(prop):
    """
    Propfan
    -------

    Fundamental difference with a propeller:
    - Disk loading.
    """



from huracan.utils import setattr_namespace


class process:
    """
    Process data class
    """
    pass


def absolute(k,
             m, t_0, p_0):
    """
    Returns the absolute temperature and pressure of a gas
    flowing at a given
    - Mach number
    - Temperature
    - Pressure

    :param k:   [-]  Specific heat ratio
    :param m:   [M]  Flow Mach number
    :param t_0: [K]  Initial temperature
    :param p_0: [Pa] Initial pressure

    :type k:    (T: float) -> float
    :type m:    float
    :type t_0:  float
    :type p_0:  float

    :return:    process instance
    """

    p   = process()

    tau = (1+(k-1)/2*m**2)
    pi  = (1+(k-1)/2*m**2)**(k/(k-1))

    t0 = t_0*tau
    p0 = p_0*pi

    setattr_namespace(p, locals())

    return p


def diffusion(mf, cp, k, m, t_0, p_0,
              eta,
              t0=None, p0=None,
              PI=None, TAU=None):
    """
    Diffusion.

    Absolute temperature is considered constant through the process.

    The absolute pressure change is determined by the total pressure ratio
    _PI_ if provided. Else, _PI_ is calculated using the isentropic
    efficiency _nu_.

    :param mf:  [kg/s] Gas mass flow                    | -> Gas characteristics
    :param cp:  [-]    Constant pressure specific heat  |
    :param k:   [-]    Specific heat ratio              |
    :param m:   [M]    Flow Mach number                 |
    :param t_0: [K]    Initial temperature              |
    :param p_0: [Pa]   Initial pressure                _|
    :param eta: [-]    Isentropic efficiency            | -> Process characteristics
    :param PI:  [-]    Pressure ratio                   |
    :param TAU: [-]    Temperature ratio               _|

    :type mf:   float
    :type cp:   float
    :type k:    float
    :type m:    float
    :type t_0:  float
    :type p_0:  float
    :type eta:  float
    :type PI:   float
    :type TAU:  float

    :return:    process instance
    """

    p   = process()

    a = absolute(k, m, t_0, p_0)

    if isinstance(TAU, type(None)):
        TAU = 1
    if isinstance(PI, type(None)):
        pi  = (1+eta*(k-1)/2*m**2)**(k/(k-1))       # Pressure -> Total pressure
        PI  = pi/a.pi

    t00 = a.t0 if isinstance(t0, type(None)) else t0
    p00 = a.p0 if isinstance(p0, type(None)) else p0

    t01 = t00*TAU
    p01 = p00*PI

    dt  = t00*(TAU-1)
    w   = cp*dt*mf

    setattr_namespace(p, locals())

    return p


def compression(mf, cp, k, t00, p00,
                eta, PI=None, TAU=None,
                ):
    """
    Adiabatic compression.

    Input pairs:
        - eta, PI
            - TAU is calculated
        - eta, TAU
            - PI is calculated

    :param mf:  [kg/s] Gas mass flow                    | -> Gas characteristics
    :param cp:  [-]    Constant pressure specific heat  |
    :param k:   [-]    Specific heat ratio              |
    :param t00: [K]    Initial total temperature        |
    :param p00: [Pa]   Initial total pressure          _|
    :param eta: [-]    Isentropic efficiency            | -> Process characteristics
    :param PI:  [-]    Pressure ratio                   |
    :param TAU: [-]    Temperature ratio               _|

    :type mf:   float
    :type cp:   float
    :type k:    float
    :type t00:  float
    :type p00:  float
    :type eta:  float
    :type PI:   float
    :type TAU:  float

    :return:    process instance
    """

    p   = process()

    assert not isinstance(PI, type(None)) or not isinstance(TAU, type(None)), \
        "Compression process: neither PI nor TAU have been defined. At least one of them must be defined."

    if isinstance(TAU, type(None)):
        # calculate TAU
        TAU = 1+1/eta*(PI**((k-1)/k)-1)
    if isinstance(PI, type(None)):
        # calculate PI
        PI = (eta*(TAU-1)+1)**(k/(k-1))

    t01 = t00*TAU
    p01 = p00*PI

    dt  = t00*(TAU-1)
    w   = cp*dt*mf

    setattr_namespace(p, locals())

    return p


def expansion(mf, cp, k, t00, p00,
              eta, PI=None, TAU=None,
              ):
    """
    Adiabatic expansion.

    Input pairs:
        - PI, eta
            - TAU is calculated
        - TAU, eta
            - PI is calculated

    :param mf:  [kg/s] Gas mass flow                    | -> Gas characteristics
    :param cp:  [-]    Constant pressure specific heat  |
    :param k:   [-]    Specific heat ratio              |
    :param t00: [K]    Initial total temperature        |
    :param p00: [Pa]   Initial total pressure          _|
    :param eta: [-]    Isentropic efficiency            | -> Process characteristics
    :param PI:  [-]    Pressure ratio                   |
    :param TAU: [-]    Temperature ratio               _|

    :type mf:   float
    :type cp:   float
    :type k:    float
    :type t00:  float
    :type p00:  float
    :type eta:  float
    :type PI:   float
    :type TAU:  float

    :return:    process instance
    """

    p   = process()

    assert not isinstance(PI, type(None)) or not isinstance(TAU, type(None)), \
        "Expansion process: neither PI nor TAU have been defined. At least one of them must be defined."

    if isinstance(TAU, type(None)):
        # calculate TAU
        TAU = 1-eta*(1-PI**((k-1)/k))
    if isinstance(PI, type(None)):
        # calculate PI
        PI = (1-1/eta*(1-TAU))**(k/(k-1))

    t01 = t00*TAU
    p01 = p00*PI

    dt  = t00*(TAU-1)
    w   = cp*dt*mf

    setattr_namespace(p, locals())

    return p


def combustion(mf, cp,
               t00, p00,
               fuel_mf, fuel_LHV,
               eta,
               PI=1):
    """
    Constant pressure heat addition.
    - A pressure ratio PI may be specified if required.

    :param mf:       [kg/s] Gas mass flow
    :param cp:       [-]    Constant pressure specific heat
    :param t00:      [K]    Initial total temperature
    :param p00:      [Pa]   Initial total pressure
    :param fuel_mf:  [kg/s] Fuel mass flow
    :param fuel_LHV: [J/kg] Fuel Lower Heating Value (heat of combustion)
    :param eta:      [-]    Isentropic efficiency

    :type mf:        float
    :type fuel_mf:   float
    :type fuel_LHV:  float
    :type t00:       float
    :type p00:       float
    :type eta:       float
    """

    p = process()

    dt = eta*(fuel_mf*fuel_LHV)/(mf*cp)

    t01 = t00 + dt
    p01 = p00*PI

    setattr_namespace(p, locals())

    return p



from alexandria.shell import print_color


def units(s, u):
    if not isinstance(s, type(str)):
        s = str(s)
    return s + ' '*(35-len(s.replace("\n", ""))) + '[{}]'.format(u.rstrip())


def result(var, val, u, r=4):
    print(units(str('{} = {:,.'+'{}'.format(r)+'f}').format(var, val), u))


def print_p_t(p, t, stage='n+1'):
    result('p0_{}'.format(stage), p, 'Pa')
    result('T0_{}'.format(stage), t, 'K')


def print_positive(text):
    print_color(text, 'yellow')


def print_negative(text):
    print_color(text, 'yellow')


def cycle_graph(x, y, title):
    import matplotlib.pyplot as plt
    from mpl_plotter.two_d import line, scatter

    line(x=x, y=y, color='grey', line_width=1, zorder=1, alpha=0.65)
    scatter(ax=plt.gca(), fig=plt.gcf(), x=x, y=y, norm=y,
            point_size=100, zorder=2, cmap='RdYlBu_r',
            x_tick_number=5, y_tick_number=5, title=title,
            x_label='p', y_label='T [K]', y_label_rotation=90, show=True)


def out(arg):
    import sys
    print_color(arg, 'red')
    sys.exit()



from tests.verification.model_verification.resources.utils_propulsion import *


def propeller(pbp, tbp, k, pi, nu, stage):
    """
    :param pbp: Total pressure ante propeller
    :param tbp: Total temperature ante propeller
    :param k: Specific heat capacity of gas flowing through propeller
    :param pi: Propeller pressure ratio
    :param nu: Propeller isentropic efficiency
    :param stage: Stage
    :return: Total pressure and temperature post propeller
    """
    p = pbp*pi
    t = tbp*(1+1/nu*(pi**((k-1)/k)-1))
    print_color('Propeller', 'green')
    print(units('T0_{} = {:,.4f}'.format(stage, t), 'K')
          +
          units('\np0_{} = {:,.4f}'.format(stage, p), 'Pa')
          )
    return p, t


def propeller_power(pbr):
    """
    Propeller required power
    :param pbr: Propeller power
    :return: Work required by the flow on turbine
    """
    print(units('Ẇ_prop = {:,.4f}'.format(pbr), 'W'))
    return pbr


def propeller_thrust(w, nu, v0):
    """
    Thrust generated by propeller
    :param w: Propeller power
    :param nu: Propeller mechanical efficiency
    :param v0: Free stream velocity
    :return:
    """
    T_prop = nu*w/v0
    print_color('Propeller thrust', 'yellow')
    print(units('T_prop = {:,.4f}'.format(T_prop), 'N'))
    return T_prop


import sys
import inspect


def setattr_namespace(o, namespace):
    """
    Set all variables declared in a namespace as as attributes
    of a class instance.
    ---

    1. Obtain list of module names
    2. Get namespace variables
       - Discard all variables with names starting and ending with '_'
    3. Create a dictionary of namespace variable names and values
    4. Set namespace variables as attributes of the input object
       - Given class instance _o_ will not set as attribute of itself
       - The parent class of _o_ will not be set as an attribute of _o_
         if present in the provided namespace.

    :param o: Instance of any given class.
    :param namespace: A given namespace.
                      - locals()
                      - globals()

    :type o: object
    :type namespace: dict
    """
    # List of module names
    _mods_ = list(set(sys.modules) & set(namespace))
    # List of function arguments
    _args_ = setattr_namespace.__code__.co_varnames
    # Get namespace variables:
    #    List of local variables which are not special variables nor module names
    keys = [key for key in namespace.keys() if (key[0] != '_' and key[-1] != '_') and key not in _mods_]
    # Dictionary of namespace variable names and values
    vars = {key: namespace[key] for key in keys}
    for key, value in vars.items():
        if not type(value) == type(o)\
                and not isinstance(o, value if inspect.isclass(value) else type(value)):  # Avoid _o_, parent of _o_
            # Set namespace variables as attributes of the input object
            setattr(o, key, value)



from tests.verification.model_verification.resources.utils_propulsion import *


def afterburner(pbab, tbab, k, pi, nu, stage):
    """
    :param pbab: Total pressure ante afterburner
    :param tbab: Total temperature ante afterburner
    :param k: Specific heat capacity of gas flowing through afterburner
    :param pi: Afterburner pressure ratio
    :param nu: Afterburner isentropic efficiency
    :param stage: Stage
    :return:
    """
    p = pbab*pi
    t = tbab*(1-nu*(1-pi**((k-1)/k)))
    print(units('T0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', t), 'K')
          +
          units('\np0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', p), 'Pa')
          )
    return p, t


def afterburner_t_given(pbab, taab, pi, stage):
    """
    :param pbab: Total pressure ante afterburner
    :param taab: Total temperature post afterburner
    :param pi: Afterburner pressure ratio
    :param stage: Stage
    :return: Total pressure and temperature post combustion chamber
    """
    p = pbab * pi
    t = taab
    print_color('Afterburner', 'green')
    print(units('T0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', t), 'K')
          +
          units('\np0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', p), 'Pa')
          )
    return p, t


def afterburner_mf(tbab, taab, mf, fmf, nu, cp, LHV):
    """
    :param tbab: Total temperature ante afterburner
    :param taab: Total temperature post afterburner
    :param mf: Gas mass flow
    :param fmf: Fuel mass flow
    :param nu: Afterburner isentropic efficiency
    :param cp: Specific heat capacity of flow through afterburner
    :param LHV: Fuel latent heat of vaporization
    :return:
    """
    abmf = (mf+fmf)*cp*(taab-tbab)/(nu*LHV)
    print_color('Afterburner fuel mass flow', 'yellow')
    print(units('SFC = {:,.4f}'.format(abmf), 'kg/s'))
    return abmf



from copy import deepcopy

from huracan.engine import component
from huracan.components import turbine


class combustor(component):
    """
    Combustor
    ---------

    Constant pressure heat addition.
    """
    def __init__(self,
                 fuel,
                 eta,
                 PI=1,
                 t01=None,
                 ):
        """
        The total pressure and temperature
        after the combustor are calculated
        based on:

        - The fuel mass flow (attribute of
          the input fuel class instance) if provided,
          if the heat added if equal or larger than
          that required by the turbines.
          -> Heat addition from input fuel mass flow

        - The total temperature of the gas
          at the exit of the combustion chamber t01
          if provided, if the heat added if equal or
          larger than that required by the turbines.
          -> Heat addition so as to reach t01

        :param fuel: Instance of fuel class.
        :param eta:  Isentropic efficiency.
        :param t01:  Temperature after the combustion chamber.

        :type fuel:  fuel
        :type t01:   float
        """

        self.fuel = fuel
        self.eta  = eta
        self.PI   = PI
        self.t01  = t01

    def tf(self, gas):
        """
        By considering combustion happens at low temperature (that
        of the gas before it enters the combustion chamber), an estimate
        of the total temperature of the gas when it exits the combustion
        chamber is obtained.
        This total temperature is then assumed to be the temperature
        at which the combustion process takes place.
        """

        assert hasattr(self, 'fmf'), \
            'Component incorporating combustor has no fuel mass flow method implemented.'

        self.fmf(gas)

        self.Q = self.fuel.mf*self.eta*self.fuel.LHV      # Heat added to the flow

        approx_process_t = deepcopy(gas).heat_addition(eta=self.eta,
                                                       cp=gas.cp(gas.t0),
                                                       fuel_mf=self.fuel.mf,
                                                       fuel_LHV=self.fuel.LHV).t01

        return gas.heat_addition(eta=self.eta,
                                 PI=self.PI,
                                 cp=gas.cp(approx_process_t),
                                 fuel_mf=self.fuel.mf,
                                 fuel_LHV=self.fuel.LHV)

    def mf_dt(self, dt, gas):
        """
        Fuel mass flow from a required change in total temperature.
        """
        gas_cp = gas.cp(self.t01)       # Heat addition happens at high temperature
        return dt*gas.mf*gas_cp/(self.eta*self.fuel.LHV - dt*gas_cp)


class combustion_chamber(combustor):
    """
    Combustion chamber
    ------------------

    Constant pressure heat addition.
    """
    def __init__(self,
                 fuel,
                 eta,
                 PI=1,
                 t01=None,
                 ):
        """
        The total pressure and temperature
        after the combustion chamber are calculated
        based on:

        - The heat required by all downstream
          turbines (work which the turbines of the
          engine must subtract from the flow).
          -> Heat addition so as to feed the known
             energy use.

        - The fuel mass flow (attribute of
          the input fuel class instance) if provided,
          if the heat added if equal or larger than
          that required by the turbines.
          -> Heat addition from input fuel mass flow

        - The total temperature of the gas
          at the exit of the combustion chamber t01
          if provided, if the heat added if equal or
          larger than that required by the turbines.
          -> Heat addition so as to reach t01

        :param fuel: Instance of fuel class.
        :param eta:  Isentropic efficiency.
        :param t01:  Temperature after the combustion chamber.

        :type fuel:  fuel
        :type t01:   float
        """
        super().__init__(fuel=fuel,
                         eta=eta,
                         PI=PI,
                         t01=t01)

    def fmf(self, gas):
        """
        At the time of execution, the fuel mass flow is
        determined based on:

        - The heat required by the turbines (minimum mass flow)
        - The fuel mass flow if provided, if larger than the
          minimum mass flow
        - The total temperature of the gas at the exit of
          the combustion chamber t01 if provided, if the
          required mass flow for it is larger than the
          minimum mass flow
        """

        mf_min = self.mf_qr(self.Q_min(downstream=self.stream.downstream))

        if not isinstance(self.t01, type(None)):
            mf_t01 = self.mf_dt(dt=self.t01 - gas.t0, gas=gas)
            self.fuel.mf = mf_t01 if mf_t01 > mf_min else mf_min
        elif not isinstance(self.fuel.mf, type(None)):
            self.fuel.mf = self.fuel.mf if self.fuel.mf > mf_min else self.fuel.mf
        else:
            self.fuel.mf = mf_min

        gas += self.fuel

    def mf_qr(self, qr):
        """
        Fuel mass flow from heat addition required.
        """
        return qr/(self.eta*self.fuel.LHV)

    def Q_min(self, downstream):
        """
        Obtain the heat required by all downstream turbines.

        Assumptions:
        - A single combustion chamber is used to power all
          downstream turbines, without secondary combustion
          chambers in any of the affluent streams.
          If this is the case, the combustion chambers will
          provide the heat required by each of their common
          turbines (and so twice the amount needed).
          For now, it is wise to ensure any secondary combustion
          chambers do not have shared turbines with the main one.

        :type downstream: list of stream

        :return: Heat required by all downstream turbines.
        """
        Qr = []
        for stream in downstream:
            for c in stream.components:
                if isinstance(c, turbine):
                    Qr.append(c.w_r())
        return sum(Qr)


class afterburner(combustor):
    """
    Afterburner or Reheater
    -----------------------

    Constant pressure heat addition.
    """
    def __init__(self,
                 fuel,
                 eta,
                 PI=1,
                 t01=None):
        """
        The total pressure and temperature
        after the afterburner are calculated
        based on:

        - The fuel mass flow (attribute of
          the input fuel class instance) if provided,
          if the heat added if equal or larger than
          that required by the turbines.
          -> Heat addition from input fuel mass flow

        - The total temperature of the gas
          at the exit of the combustion chamber t01
          if provided, if the heat added if equal or
          larger than that required by the turbines.
          -> Heat addition so as to reach t01

        :param fuel: Instance of fuel class.
        :param eta:  Isentropic efficiency.
        :param t01:  Temperature after the combustion chamber.

        :type fuel:  fuel
        :type t01:   float
        """
        super().__init__(fuel=fuel,
                         eta=eta,
                         PI=PI,
                         t01=t01)

    def fmf(self, gas):
        """
        At the time of execution, the fuel mass flow is
        determined based on:

        - The fuel mass flow if provided, if larger than the
          minimum mass flow
        - The total temperature of the gas at the exit of
          the combustion chamber t01 if provided, if the
          required mass flow for it is larger than the
          minimum mass flow
        """
        assert not isinstance(self.t01, type(None)) or not isinstance(self.fuel.mf, type(None)), \
            'Afterburner: neither the fuel mass flow nor the temperature t01 after the afterburner have been' \
            'provided. Input either one in the initialization of your afterburner instance to proceed.'

        if not isinstance(self.t01, type(None)):
            self.fuel.mf = self.mf_dt(dt=self.t01 - gas.t0, gas=gas)
        elif not isinstance(self.fuel.mf, type(None)):
            self.fuel.mf = self.fuel.mf

        gas += self.fuel



import unittest

from huracan.utils import setattr_namespace


class TestsFluid(unittest.TestCase):

    def test_setattr_namespace(self):
        class a:
            pass

        print(type(a))

        b = a()

        c = 1

        setattr_namespace(b, locals())




from alexandria.shell import print_color
from alexandria.data_structs.string import join_set_distance


def verify(prediction, measurement, error_margin=0.001, log=False, name='Error'):
    """
    Model verification:
    - Error must stay under 0.1% of the measurement.

    :param error_margin: [>= 0] Error margin in unit fractions (1 ~ 100%).

    :type prediction:    float
    :type measurement:   float
    :type error_margin:  float
    :type log:           bool
    """
    assert abs(prediction-measurement) < measurement*error_margin, \
        f'Error = {abs(prediction-measurement)/measurement*100:.2f}% of measurement.'

    if log:
        s = join_set_distance(name, '=', 10)
        print_color(s + f' {abs(prediction-measurement)/measurement*100:.2f}% of measurement.', 'yellow')



from tests.verification.model_verification.resources.functions_propulsion import *
from tests.verification.model_verification.resources.utils_propulsion import *


def pi_compression(pi, nu, k, p, t):
    print_positive('Compression')
    p = pi*p
    t = t*(1+1/nu*(pi**((k-1)/k)-1))
    print_p_t(p, t)
    return p, t


def pi_expansion(pi, nu, k, p, t):
    print_positive('Expansion')
    p = pi*p
    t = t*(1-nu*(1-pi**((k-1)/k)))
    print_p_t(p, t)
    return p, t


def dt_expansion(pb, tb, dt, nu, k):
    print_positive('Expansion')
    t = tb - dt
    p = pb*(1-1/nu*(1-t/tb))**(k/(k-1))
    print_p_t(p, t)
    return p, t


def work_done(dt, mf, cp):
    print_positive('Work exerted')
    w_c = dt*mf*cp
    result('W_C', w_c, 'J')
    return w_c


def work_required(w_c, nu):
    print_positive('Work required')
    w_r = w_c/nu
    result('W_R', w_r, 'J')
    return w_r


def work_dt(w, mf, cp):
    print_positive('Work'+'\u0394'+'T')
    dt = w/mf/cp
    result('\u0394'+'T', dt, 'J')
    return dt


def nozzle(pbn, pamb, tbn, tan, tamb, nu_nozz, nu, mf, m, k=1.33, cp=1150, R=287, which='Core'):
    """
    :return: Nozzle results
    """

    "Choked nozzle"
    p_o_pcrit, choked = is_nozzle_choked(pbn/pamb, nu_nozz, k)

    if choked is True:
        "Pressure, temperature"
        print_positive('Nozzle'.format(which))
        p = pbn / p_o_pcrit
        pi = p / pbn
        t = tbn*(1-nu*(1-pi**((k-1)/k)))
        print_p_t(p, t)
        "Flow exit velocity"
        print_positive('{} exit velocity'.format(which))
        v = (2 * cp * (tbn - tan)) ** 0.5
        result('v{}'.format('8' if which == 'Core' else '18'), v, 'm/s')
    else:
        "Pressure, temperature"
        print_positive('Nozzle'.format(which))
        p = pamb
        pi = pamb / pbn
        t = tbn*(1-nu*(1-pi**((k-1)/k)))
        print_p_t(p, t)
        "Flow exit velocity"
        print_positive('{} exit velocity'.format(which))
        v = (k*R*tan)**0.5
        result('v{}'.format('8' if which == 'Core' else '18'), v, 'm/s')

    "Flow thrust"
    fn = mf*(v-v0(k, R, tamb, m))+A_exit(mf, t, p, v, R, 'n+1')*(p-pamb)
    result('T', fn, 'N')

    return p, t, v, fn



"""
Huracan
-------
Twin-spool, reheated turbojet engine test.
"""

from huracan.engine import shaft
from huracan.thermo.fluids import gas, fuel
from huracan.components import inlet, compressor, combustion_chamber, turbine, afterburner, nozzle

from tests.utils import verify

mf = 160
M  = 0
t  = 288
p  = 101325

fuel_cc = fuel(LHV=43e6)
fuel_ab = fuel(LHV=43e6)

g = gas(mf=mf,
        cp=lambda T: 1150 if T > 600 else 1000,
        k=lambda T: 1.33 if T > 600 else 1.4,
        m=M, t_0=t, p_0=p)

i  = inlet(PI=0.92)
c1 = compressor(eta=0.85, PI=4)
c2 = compressor(eta=0.85, PI=4)
cc = combustion_chamber(fuel_cc, eta=0.97, t01=1450)
t1 = turbine(0.9)
t2 = turbine(0.9)
ab = afterburner(fuel_ab, eta=0.95, t01=1850)
n  = nozzle(0.95)

shaft1 = shaft(c1, t2, eta=0.99)
shaft2 = shaft(c2, t1, eta=0.99)

stream = g-i-c1-c2-cc-t1-t2-ab-n

stream.run()

"""
Verification

- Error must stay below 0.1% of the verification value 
"""
# Total temperature
verify(stream['0.il'].t0,    288)               # Inlet
verify(stream['0.cp1'].t0, 452.66630032)      # Compressor 1
verify(stream['0.cp2'].t0, 711.48187307)      # Compressor 2

# Total pressure
verify(stream['0.il'].p0,    93219)             # Inlet
verify(stream['0.cp1'].p0, 372876)           # Compressor 1
verify(stream['0.cp2'].p0, 1491504)           # Compressor 2

"""
Exceptions:

The fuel mass flow method differs from
that used by the verification model. 
This causes differences in the models'
results after the main combustion chamber.

- Error must stay below 5% of the verification value 
"""
# Fuel mass flow
verify(cc.fuel.mf, 3.1921, 0.05)

# Total temperature
verify(stream['0.tb1'].t0, 1227.1163336,  0.05)    # Turbine 1
verify(stream['0.tb2'].t0, 1085.31099956, 0.05)    # Turbine 2

# Total pressure
verify(stream['0.tb1'].p0, 680119.41922727, 0.05)  # Turbine 1
verify(stream['0.tb2'].p0, 390882.21425195, 0.05)  # Turbine 2

# Exit velocity
verify(stream.v_exit(), 840.3353)



import numpy as np
from tests.verification.model_verification.turbojet import Turbojet
from tests.verification.model_verification.resources.utils_propulsion import cycle_graph

tj = Turbojet()

tj.h = 0
tj.M = 0
tj.mf = 160
tj.T0_4 = 1450
tj.T0_7 = 1850
tj.T_amb = 288.00
tj.p_amb = 101325.00

# pressure ratios
tj.pi_inlet = 0.92
tj.pi_LPC = 4
tj.pi_HPC = 4
tj.pi_cc = 0.97
tj.pi_ab = 0.97

# isentropic efficiencies
tj.nu_inlet = 1.00
tj.nu_LPC = 0.85
tj.nu_HPC = 0.85
tj.nu_LPT = 0.9
tj.nu_HPT = 0.9
tj.nu_mech = 0.99
tj.nu_cc = 0.99
tj.nu_nozzle = 0.95
tj.nu_ab = 0.95

# constants
tj.LHV = 43e6
tj.cp_air = 1000
tj.cp_gas = 1150
tj.k_air = 1.4
tj.k_gas = 1.33
tj.R = 287

tj.ambient()
tj.inlet()
tj.lpc('25')
tj.hpc('3')
tj.cc('4')
tj.fmf()
tj.work_exerted()
tj.work_required()
tj.hpt('45')
tj.lpt('5')
tj.afterburner('7')
tj.abmf()
tj.pi_nozzle_core()
tj.nozzle_core('8')
tj.exit_velocity()
tj.nozzle_exit_area('8')
tj.thrust_core()
tj.sfc()

x, y = zip(*tj.record)
x, y = np.array(x), np.array(y)

cycle_graph(x, y, 'Turbojet')





from huracan.engine import component


class intake(component):
    """
    Intake
    ------

    Airflow fed directly to engine.
    """
    def __init__(self,
                 eta=1,
                 PI=1,
                 TAU=1):
        """
        :type eta: float
        :type PI:  float
        :type TAU: float
        """
        self.eta = eta
        self.PI  = PI
        self.TAU = TAU

    def tf(self, gas):
        return gas.diffusion(eta=self.eta, PI=self.PI, TAU=self.TAU)


class inlet(component):
    """
    Inlet
    -----

    Adiabatic diffusion.
    """
    def __init__(self,
                 eta=None,
                 PI=None,
                 TAU=None):
        """
        :type eta: float
        :type PI:  float
        :type TAU: float
        """
        self.eta = eta
        self.PI  = PI
        self.TAU = TAU

        assert not isinstance(eta, type(None)) or not isinstance(PI, type(None)) or not isinstance(TAU, type(None))

    def tf(self, gas):
        return gas.diffusion(eta=self.eta, PI=self.PI, TAU=self.TAU)


class nozzle(component):
    """
    Nozzle
    ------

    Adiabatic expansion.
    """
    def __init__(self,
                 eta,
                 PI=None
                 ):
        """
        :type eta: float
        :type PI:  float
        """
        self.eta = eta
        self.PI  = PI

        self.choked = False

    def tf(self, gas):
        return gas.expansion(eta=self.eta, PI=self.pi(gas), TAU=None)

    def pi(self, gas):
        """
        Nozzle pressure ratio.

        If the nozzle pressure ratio is not provided, it is set as
        the ratio between the gas total pressure before the nozzle
        and the ambient pressure.
        If the inverse of the nozzle pressure ratio is larger than
        the critical inverse pressure ratio, the flow is choked and
        the nozzle pressure ratio will not exceed the critical pressure
        ratio.
        """
        inv_PI = gas.p0/gas.p_0 if isinstance(self.PI, type(None)) else 1/self.PI

        if inv_PI > self.inv_pi_crit(gas):
            PI = 1/self.inv_pi_crit(gas)
            self.choked = True
        else:
            PI = 1/inv_PI

        return PI

    def inv_pi_crit(self, gas):
        """
        Inverse critical pressure ratio.

        Ratio of the pressure **before** the inlet over that
        **after** the inlet (inversely to the widely used pressure
        ratio), starting from which supersonic flow happens in the
        expansion process, causing the nozzle to become choked.

        With a choked nozzle the expansion of the gas is incomplete,
        with a maximum total pressure ratio in the nozzle equal to

            1/PI_crit

        """
        k = gas.k(gas.t0)
        return 1/(1-1/self.eta*((k-1)/(k+1)))**(k/(k-1))



from tests.verification.model_verification.resources.utils_propulsion import *


def ambient(pamb, tamb, m, k):
    """
    :param pamb: Ambient pressure
    :param tamb: Ambient temperature
    :param m: Mach number
    :param k: Specific heat capacity of gas flowing through inlet
    :return: Total ambient pressure and temperature
    """
    print_color('Ambient', 'green')
    p = pamb*(1+(k-1)/2*m**2)**(k/(k-1))
    t = tamb*(1+(k-1)/2*m**2)
    print(units('t0_0 = {:,.4f}'.format(t), 'K')
          +
          units('\np0_0 = {:,.4f}'.format(p), 'Pa')
          )
    return p, t


def inlet(pamb, tamb, m, k, pi, nu):
    """
    :param pamb: Ambient pressure
    :param tamb: Ambient temperature
    :param m: Mach number
    :param k: Specific heat capacity of gas flowing through inlet
    :param pi: Inlet pressure ratio
    :param nu: Inlet isentropic efficiency
    :return: Total pressure and temperature after inlet
    """
    p = pi*pamb*(1+(k-1)/2*m**2)**(k/(k-1))
    t = tamb*(1+(k-1)/2*m**2)
    p_nu = pi*pamb*(1+nu*(k-1)/2*m**2)**(k/(k-1))
    print_color('Inlet', 'green')
    if nu != 1:
        print(units('t0_2 = {:,.4f}'.format(t), 'K')
              +
              units('\np0_2 = {:,.4f} [nu = 1]'.format(p), 'Pa')
              +
              units('\np0_2 = {:,.4f} [nu = {}]'.format(p_nu, nu), 'Pa'))
    else:
        print(units('t0_2 = {:,.4f}'.format(t), 'K')
              +
              units('\np0_2 = {:,.4f}'.format(p), 'Pa')
              )
    return p, t


def compressor(pbc, tbc, k, pi, nu, stage=None, kind=None):
    """
    Total pressure and temperature through compressor stage
    :param pbc: Total pressure ante compressor
    :param tbc: Total temperature ante compressor
    :param k: Specific heat capacity of gas flowing through compressor
    :param pi: Compressor pressure ratio
    :param nu: Compressor isentropic efficiency
    :param stage: Stage
    :param kind: HPC, LPC
    :return: Total pressure and temperature post compressor stage
    """
    p = pi*pbc
    t = tbc*(1+1/nu*(pi**((k-1)/k)-1))
    print_color('{}'.format(kind), 'green')
    print(units('T0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', t), 'K')
          +
          units('\np0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', p), 'Pa')
          )
    return p, t


def turbine(w, mf, cp, k, nu, tbt, pbt, stage=None, kind=None):
    """
    Total pressure and temperature through turbine stage
    :param w: Work done by flow on turbine
    :param mf: ṁ through turbine
    :param cp: Specific heat capacity of flow through turbine
    :param k: Specific heat capacity of gas flowing through compressor
    :param nu: Turbine isentropic efficiency
    :param tbt: Total temperature ante turbine
    :param pbt: Total pressure ante turbine
    :param stage: Stage
    :param kind:
    :return: Total pressure and temperature post turbine stage
    """
    t = tbt - w/(mf*cp)
    p = pbt*(1-1/nu*(1-t/tbt))**(k/(k-1))
    print_color('{}'.format(kind), 'green')
    print(units('T0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', t), 'K')
          +
          units('\np0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', p), 'Pa')
          )
    return p, t


def compr_work(mf, cp, tbc, tac, which='LPC'):
    """
    Work done by compressor
    :param mf: ṁ through compressor
    :param cp: Specific heat capacity of gas through compressor
    :param tbc: T ante compressor stage
    :param tac: T post compressor stage
    :param which: LPC or HPC
    :return: Worked done by compressor
    """
    w_comp = mf*cp*(tac-tbc)
    print_color('{} work on flow'.format(which), 'yellow')
    print(units('Ẇ_{} = {:,.4f}'.format(which.lower(), w_comp), 'W'))
    return w_comp


def turbine_pr(w_comp, nu, which='LPT'):
    """
    Turbine power required
    :param w_comp: work exerted by compressor in shared spool
    :param nu_spool: Spool mechanical efficiency
    :param which: LPC or HPC
    :return: Power required from turbine
    """
    p_turb = w_comp/nu
    print_color('{} power required'.format(which), 'yellow')
    print(units('Ẇ_{} = {:,.4f}'.format(which, p_turb), 'W'))
    return p_turb


def turbine_dt(w, mf, cp):
    """
    Change of temperature due to work done by flow to power turbine in turbine stage
    :param w: Work done by flow on turbine
    :param mf: Mass flow through turbine
    :param cp: Specific heat capacity of flow through turbine
    :return: Turbine ∆T
    """
    return print(units('∆T = {:,.4f}'.format(-w/(mf*cp)), '°C'))


def fuel_mass_flow(tbc, tac, mf, cp, nu, lhv):
    """
    Fuel mass flow
    :param tbc: Temperature ante combustion
    :param tac: Temperature post combustion
    :param mf: ṁ through combustion chamber/afterburner
    :param cp: Specific heat capacity of hot gas through compressor
    :param nu: Combustion chamber/afterburner isentropic efficiency
    :param lhv: Fuel latent heat of vaporization
    :return: Fuel ṁ
    """
    dt = tac-tbc
    fmf = mf*cp*dt/(lhv*nu)
    print_color('Fuel mass flow', 'yellow')
    print(units('Fuel ṁ = {:,.4f}'.format(fmf), 'Kg/s'))
    return fmf


def cc_t_given(pbcc, tacc, pi, stage):
    """
    :param pbcc: Total pressure ante combustion chamber
    :param tacc: Total temperature post combustion chamber
    :param pi: Combustion chamber pressure ratio
    :param stage: Stage
    :return: Total pressure and temperature post combustion chamber
    """
    p = pbcc*pi
    print_color('Combustion chamber', 'green')
    print(units('T0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', tacc), 'K')
          +
          units('\np0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', p), 'Pa')
          )
    return p, tacc


def cc_dt_given(pbcc, tbcc, dt, pi, stage):
    """
    :param pbcc: Total pressure ante combustion chamber
    :param tbcc: Total temperature after combustion chamber
    :param dt: Total temperature gradient after combustion chamber
    :param pi: Combustion chamber pressure ratio
    :param stage: Stage
    :return: Total pressure and temperature post combustion chamber
    """
    p = pbcc*pi
    t = tbcc + dt
    print_color('Combustion chamber', 'green')
    print(units('T0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', t), 'K')
          +
          units('\np0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', p), 'Pa')
          )
    return p, t


def cc_fmf_given(pbcc, tbcc, pi, mf, fmf, nu, cp, LHV, stage):
    """
    :param pbcc: Total pressure ante combustion chamber
    :param tbcc: Total temperature after combustion chamber
    :param pi: Combustion chamber pressure ratio
    :param mf: Air mass flow
    :param fmf: Fuel mass flow
    :param nu: Combustion chamber isentropic efficiency
    :param cp: Specific heat capacity of hot gas through combustion chamber
    :param LHV: Fuel latent heat of vaporization
    :param stage: Stage
    :return: Total pressure and temperature post combustion chamber
    """
    p = pbcc*pi
    t = tbcc + nu*fmf*LHV/(mf*cp)
    print_color('Combustion chamber', 'green')
    print(units('T0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', tbcc), 'K')
          +
          units('\np0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', p), 'Pa')
          )
    return p, t


def cc_t_interturbine(pbcc, t_inter, pi, w_comp, mf, cp, nu, stage):
    """
    :param pbcc: Total pressure before combustion chamber
    :param t_inter: Total interturbine temperature
    :param pi: Combustion chamber pressure ratio
    :param w_comp: HPC work on flow
    :param mf: Mass flow through combustion chamber
    :param cp: Specific heat capacity of hot gas through combustion chamber
    :param nu: Combustion chamber isentropic efficiency
    :param stage: Stage
    :return: Total pressure and temperature at combustion chamber exit
    """
    p = pbcc*pi
    t = t_inter + w_comp/mf/cp/nu
    print_color('Combustion chamber', 'green')
    print(units('T0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', t), 'K')
          +
          units('\np0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', p), 'Pa')
          )
    return p, t


def cc_w_given(pbcc, tbcc, pi, w_HPT, w_LPT, mf, cp, nu, stage):
    """
    :param pbcc: Total pressure before combustion chamber
    :param tbcc: Total temperature after combustion chamber
    :param pi: Combustion chamber pressure ratio
    :param w_HPT: HPT work required
    :param w_LPT: LPT work required
    :param mf: Mass flow through combustion chamber
    :param cp: Specific heat capacity of hot gas through combustion chamber
    :param nu: Combustion chamber isentropic efficiency
    :param stage: Stage
    :return: Total pressure and temperature at combustion chamber exit
    """
    p = pbcc*pi
    t = tbcc + (w_HPT+w_LPT)/mf/cp
    t_nu = tbcc + (w_HPT+w_LPT)/mf/cp/nu
    print_color('Combustion chamber', 'green')
    if nu != 1:
        print(units('T0_{} = {:,.4f} [nu = 1]'.format(stage if not isinstance(stage, type(None)) else 'n+1', t), 'K')
              +
              units('\nT0_{} = {:,.4f} [nu = {}]'.format(stage if not isinstance(stage, type(None)) else 'n+1', t_nu, nu), 'K')
              +
              units('\np0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', p), 'Pa')
              )
    else:
        print(units('T0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', t), 'K')
              +
              units('\np0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', p), 'Pa')
              )
    return p, t


def pi_nozzle(p_nozzle, p_amb, which):
    """
    :param p_nozzle: Pressure before nozzle exit
    :param p_amb: Ambient pressure
    :param which: Which
    :return: Nozzle pressure ratio
    """
    pi = p_nozzle/p_amb
    print_color('{} pressure ratio'.format(which), 'yellow')
    print(units('pi_{} = {:,.4f}'.format(which.lower(), pi), '-'))
    return pi


def is_nozzle_choked(pi_nozz, nu_nozz, k):
    """
    Determining whether the nozzle is choked:
        Choked:
            total pressure after nozzle will equal pcrit = (p/p_crit)^-1 * p
        Non-choked:
            total pressure after nozzle will equal p
    :param pi_nozz: Nozzle expansion ratio
    :param nu_nozz: Nozzle isentropic efficiency
    :param k: Specific heat capacity of gas flowing through nozzle
    :return: Nozzle condition
    """
    p_pcrit = 1/(1-1/nu_nozz*((k-1)/(k+1)))**(k/(k-1))
    choked = pi_nozz > p_pcrit
    print_color('{} nozzle '.format('Choked' if choked is True else 'Non choked'),
                  'red' if pi_nozz > p_pcrit else 'green')
    print('    p/p_crit = {}'.format(p_pcrit))
    return p_pcrit, choked


def non_choked_nozzle(pbn, tbn, pamb, k, nu, stage=None):
    """
    Total pressure and temperature through non-choked nozzle
    :param pbn: Total pressure ante nozzle
    :param tbn: Total temperature ante nozzle
    :param pamb: Ambient pressure
    :param k: Specific heat capacity of gas flowing through nozzle
    :param nu: Nozzle isentropic efficiency
    :param stage: Stage
    :return: Total pressure and temperature post non-choked nozzle stage
    """
    pi = pamb/pbn
    t = tbn*(1-nu*(1-pi**((k-1)/k)))
    print(units('T{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', t), 'K')
          +
          units('\np{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', pamb), 'Pa')
          )
    return pamb, t


def choked_nozzle(pbn, tbn, p_pc, k, nu, stage=None):
    """
    Total pressure and temperature through choked nozzle
    :param pbn: Total pressure ante nozzle
    :param tbn: Total temperature ante nozzle
    :param p_pc: Total pressure post combustion chamber
    :param k: Specific heat capacity of gas flowing through nozzle
    :param nu: Nozzle isentropic efficiency
    :param stage: Stage
    :return: Total pressure and temperature post choked nozzle stage
    """
    p = pbn/p_pc
    pi = p/pbn
    t = tbn*(1-nu*(1-pi**((k-1)/k)))
    print(units('T{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', t), 'K')
          +
          units('\np{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', p), 'Pa')
          )
    return p, t


def exit_velocity_choked(k, R, tan, which):
    """
    Choked exit flow velocity
    :param k: Specific heat capacity of gas flowing through nozzle
    :param R: Ideal gas constant (287)
    :param tan: Total temperature after nozzle
    :param which: Core or bypass
    :return: Flow velocity at nozzle exit -core flow velocity-
    """
    v = (k * R * tan) ** 0.5
    den = '8' if which == 'Core' else '18'
    print_color('{} exit velocity'.format(which), 'yellow')
    print(units('v{} = {:,.4f}'.format(den, v), 'm/s'))
    return v


def exit_velocity_unchoked(cp, tbn, tan, which):
    """
    Unchoked exit flow velocity
    :param cp: Specific heat capacity of hot gas through nozzle
    :param tbn: Total temperature before nozzle
    :param tan: Total temperature after nozzle
    :param which: Core or bypass
    :return: Flow velocity at nozzle exit -core flow velocity-
    """
    v = (2*cp*(tbn-tan))**0.5
    den = '8' if which == 'Core' else '18'
    print_color('{} exit velocity'.format(which), 'yellow')
    print(units('v{} = {:,.4f}'.format(den, v), 'm/s'))
    return v


def A_exit(mf, t_exit, p_exit, v_exit, R, stage):
    """
    :param mf: Mass flow through nozzle
    :param t_exit: Temperature at exit of fan nozzle
    :param p_exit: Pressure at fan nozzle exit
    :param v_exit: Flow velocity at exit of fan nozzle
    :param R: Ideal gas constant (287)
    :param stage: Stage
    :return: Fan nozzle equivalent area
    """
    A_exit = mf * R * t_exit / (p_exit * v_exit)
    print_color('A{}'.format(stage), 'yellow')
    print(units('A{} = {:,.4f}'.format(stage, A_exit), 'm^2'))
    return A_exit


def flow_thrust(mf, van, v0, A_exit, pan, p0, choked, which='Core'):
    """
    Thrust generated by jet
    :param mf: ṁ through
    :param van: Flow velocity at nozzle exit
    :param v0: Free stream velocity
    :param A_exit: Nozzle exit area
    :param pan: Pressure at nozzle exit
    :param p0: Total pressure at inlet
    :param choked: Whether nozzle is choked or not
    :param which: Core or bypass
    :return: Thrust generated by jet
    """
    if choked is True:
        fn = mf*(van - v0) + A_exit*(pan - p0)
    else:
        fn = mf*(van - v0)
    print_color('{} thrust'.format(which), 'yellow')
    print(units('T_{} = {:,.4f}'.format(which, fn), 'N'))
    return fn


def sfc(fmf, T_total):
    """
    :param fmf: Fuel mass flow
    :param T_total: Total thrust
    :return: Specific fuel consumption
    """
    sfc = fmf/(T_total)
    print_color('Specific fuel consumption', 'yellow')
    print(units('SFC = {:,.4f}E-5'.format(sfc*10e4), 'kg/s/N'))
    return sfc


def intercooler_t_given(pbic, taic, pi, stage):
    """
    :param pbic: Total pressure before intercooler
    :param taic: Total temperature after intercooler
    :param pi: Intercooler pressure ratio
    :return: Total temperature and pressure after intercooler
    :param stage: Stage
    """
    p, t = pbic*pi, taic
    print_color('Intercooler', 'green')
    print(units('T0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', t), 'K')
          +
          units('\np0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', p), 'Pa')
          )
    return p, t


def intercooler_f(E_intercooler, hmf, cmf,
                  tbic_core, taic_core,
                  tbic_bp, taic_bp):
    """
    :param E_intercooler: Intercooler efficiency
    :param hmf: Hot (core) mass flow
    :param cmf: Cold (bypass) mass flow
    :param tbic_core: Core flow total temperature before intercooler
    :param taic_core: Core flow total temperature after intercooler
    :param tbic_bp: Bypass flow total temperature before intercooler
    :param taic_bp: Bypass flow total temperature after intercooler
    :param cp: Specific heat capacity of hot gas through recuperator
    :return: Bypass air fraction needed
    """
    f = hmf*(tbic_core-taic_core)/cmf/(taic_bp-tbic_bp)/E_intercooler
    print_color('Intercooler bypass flow fraction', 'yellow')
    print(units('f = {:,.4f}'.format(f), '-'))
    return f


def recuperator_heat(E_recuperator, tbr_core, dt, cp_air, cp_gas,
                     pbr_core, pi_rec_comp, stage):
    """
    :param E_recuperator: Recuperator efficiency
    :param tbr_core: Core flow total temperature before recuperator heat addition
    :param dt: Total temperature gradient due to recuperator stage -before nozzle-
    :param cp_air: Specific heat capacity of bypass (cold) gas through recuperator
    :param cp_gas: Specific heat capacity of core (hot) gas through recuperator
    :param stage: Stage
    :return: Total core temperature before combustion chamber
    """
    tar_cc = tbr_core + E_recuperator*cp_gas*dt/cp_air
    par_cc = pbr_core*pi_rec_comp
    print_color('Recuperator heat addition', 'green')
    print(units('T0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', tar_cc), 'K')
          +
          units('\np0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', par_cc), 'Pa')
          )
    return par_cc, tar_cc


def recuperator_remove(tbr_nozzle, dt, pbr_nozzle, pi_rec_nozzle, stage):
    """
    :param tbr_nozzle: Core total temperature before recuperator heat removal
    :param dt: Recuperator temperature gradient
    :param stage: Stage
    :return: Core total temperature after recuperator heat removal
    """
    tar_nozzle = tbr_nozzle - dt
    par_nozzle = pbr_nozzle*pi_rec_nozzle
    print_color('Recuperator heat removal', 'green')
    print(units('T0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', tar_nozzle), 'K')
          +
          units('\np0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', par_nozzle), 'Pa')
          )
    return par_nozzle, tar_nozzle


def hp(p):
    return p*0.735499*1000


def fuel_to_air_real(fa_real, fa_stoich):
    if fa_real > fa_stoich:
        print_color('Rich combustion', 'red')
    else:
        print_color('Lean combustion', 'red')


def compression_pi(pbc, tbc, k, pi, nu):
    p = pi*pbc
    t = tbc*(1+1/nu*(pi**((k-1)/k)-1))
    return p, t


def compression_tr(pbc, tbc, k, tr, nu):
    t = tbc*tr
    p = pbc*(1+nu*(t/tbc-1))**(k/(k-1))
    return p, t


def expansion_pi(pbe, tbe, k, pi, nu):
    p = pbe*pi
    t = tbe*(1-nu*(1-pi**((k-1)/k)))
    return p, t


def expansion_tr(pbe, tbe, k, tr, nu):
    t = tbe*tr
    p = pbe*(1+nu*(t/tbe-1))**(k/(k-1))
    return p, t


def mach(v0, k, r, t):
    return v0/(k*r*t)**0.5


def v0(k, r, t, m):
    """
    :param k: Specific heat capacity of free stream air
    :param r: Gas constant = 287J/Kg/K
    :param t: Ambient temperature
    :param m: Free stream Mach number
    :return: Free stream flow velocity
    """
    return m*(k*r*t)**0.5


import numpy as np
from tests.verification.model_verification.turboprop import Turboprop
from tests.verification.model_verification.resources.utils_propulsion import cycle_graph

tp = Turboprop()

tp.M = 0.5
tp.mf = 35
tp.Pbr = 3.91e6
tp.T0_4 = 1130
tp.T_amb = 288.15
tp.p_amb = 101325.00

# pressure ratios
tp.pi_inlet = 1
tp.pi_C = 11.5
tp.pi_cc = 0.96

# isentropic efficiencies
tp.nu_prop = 0.9
tp.nu_C = 0.85
tp.nu_T = 0.89
tp.nu_mech = 0.99
tp.nu_cc = 0.995
tp.nu_comb = 0.995
tp.nu_nozzle = 0.95

# constants
tp.LHV = 43e6
tp.cp_air = 1000
tp.cp_gas = 1150
tp.k_air = 1.4
tp.k_gas = 1.33
tp.R = 287

tp.ambient()
tp.prop('1')
tp.inlet()
tp.comp('3')
tp.cc('4')
tp.fmf()
tp.work_exerted_monocomp('')
tp.work_required_monoturb()
tp.turb('5')
tp.pi_nozzle_core()
tp.nozzle_core('8')
tp.exit_velocity()
tp.nozzle_exit_area('8')
tp.thrust_core()
tp.thrust_prop()
tp.sfc()

x, y = zip(*tp.record)
x, y = np.array(x), np.array(y)

# cycle_graph(x, y, 'Turboprop')



import unittest

from huracan.components import inlet, compressor, turbine, nozzle
from huracan.thermo.fluids import gas

mf = 700
m = 0.6
t = 288
p = 101325
fr = 0

g = gas(mf=mf,
        cp=lambda T: 1150 if T > 600 else 1000,
        k=lambda T: 1.33 if T > 600 else 1.4,
        m=m, t_0=t, p_0=p)


class TestsFluid(unittest.TestCase):

    def test_linear(self):
        i = inlet(0.95)
        c1 = compressor(0.95, 2)
        c2 = compressor(0.95, 14)

        c = g - i - c1 - c2

        c.run()

        assert c.s00.process.t0 == c.s01.process.t01        # Inlet - no change in total temperature
        assert c.s1.process.p01 == 2 * c.s01.process.p01    # Compressor 1
        assert c.s2.process.p01 == 14 * c.s1.process.p01    # Compressor 2



"""
Huracan
-------
Three-spool turbofan engine test.
"""

from huracan.engine import shaft
from huracan.thermo.fluids import gas, fuel
from huracan.components import inlet, fan, compressor, combustion_chamber, turbine, afterburner, nozzle, power_plant

from tests.utils import verify

mf = 1440
M  = 0.4
t  = 281.65
p  = 89874

bpr = 9.6

fuel = fuel(LHV=43e6)

g = gas(mf=mf,
        cp=lambda T: 1150 if T > 1000 else 1000,
        k=lambda T: 1.33 if T > 1000 else 1.4,
        m=M, t_0=t, p_0=p)

i  = inlet             (PI=0.98)
fn = fan               (eta=0.94,  PI=1.54)
c1 = compressor        (eta=0.991, PI=9.61)
c2 = compressor        (eta=0.92,  PI=3.38)
cc = combustion_chamber(fuel=fuel, eta=0.985, PI=0.99, t01=1838)
t1 = turbine           (eta=0.96)
t2 = turbine           (eta=0.965)
t3 = turbine           (eta=0.97)
nc = nozzle            (eta=0.95)
nf = nozzle            (eta=0.96)

shaft1 = shaft(fn, t3, eta=0.995)
shaft2 = shaft(c1, t2, eta=0.995)
shaft3 = shaft(c2, t1, eta=0.995)

stream = g-i-fn

s1core, s1bypass = stream*(bpr/(bpr+1))

s1core-c1-c2-cc-t1-t2-t3-nc
s1bypass-nf

stream.run()

"""
Verification

- Error must stay below 0.1% of the verification value 
"""
# Total temperature
verify(stream['0.il'].t0,    290.7)              # Inlet
verify(stream['0.fn'].t0,    331.3)              # Fan
verify(stream['1.m.nz'].t0,  287.2)              # Bypass flow nozzle
verify(stream['1.s.cp1'].t0, 635.1)              # Compressor 1
verify(stream['1.s.cp2'].t0, 922.4)              # Compressor 2

# Total pressure
verify(stream['0.il'].p0,    98343)              # Inlet
verify(stream['0.fn'].p0,    151447)             # Fan
verify(stream['1.m.nz'].p0,  89874)              # Bypass flow nozzle
verify(stream['1.s.cp1'].p0, 1455410)            # Compressor 1
verify(stream['1.s.cp2'].p0, 4919286)            # Compressor 2

# Heat and work
verify(fn.w, 58.46e6)                            # Work: Fan
verify(c1.w, 41.27e6)                            # Work: Compressor 1
verify(c2.w, 39.03e6)                            # Work: Compressor 2

"""
Exceptions:

The fuel mass flow method differs from
that used by the verification model. 
This causes differences in the models'
results after the main combustion chamber.

- Error must stay below 5% of the verification value 
"""
# Fuel mass flow
verify(cc.fuel.mf, 3.377, 0.05)

# Total temperature
verify(stream['1.s.tb1'].t0, 1593,    0.05)       # Turbine 1
verify(stream['1.s.tb2'].t0, 1334,    0.05)       # Turbine 2
verify(stream['1.s.tb3'].t0, 967,     0.05)       # Turbine 3

# Total pressure
verify(stream['1.s.tb1'].p0, 2666269, 0.05)       # Turbine 1
verify(stream['1.s.tb2'].p0, 1267264, 0.05)       # Turbine 2
verify(stream['1.s.tb3'].p0, 330414,  0.05)       # Turbine 3

# Thrust
verify(s1core._flow_thrust(), 85703, 0.1,
       True, 'e(T_core)')                         # Thrust

# Efficiency
verify(stream.sfc(), 11.36e-6, 0.1,
       True, 'e(SFC)')                            # Specific fuel consumption



"""
Huracan engine components
"""

from huracan.components.static import intake, inlet, nozzle
from huracan.components.rotary import fan, prop, propfan, compressor, turbine
from huracan.components.combustion import combustion_chamber, afterburner
from huracan.components.electrical import power_plant



from huracan.engine import component


class power_plant(component):
    """
    Aircraft electrical power plant
    -------------------------------
    """
    def __init__(self,
                 w,
                 eta_g=1,
                 eta_c=1):
        """
        :param w:     [W] Power required by the electrical power plant.
        :param eta_g: [-] Efficiency of the electrical generator attached
                          to the turbine.
        :param eta_c: [-] Efficiency of the electrical power distribution
                          grid of the aircraft.
        """
        self.w_r = w/eta_g/eta_c



from tests.verification.model_verification.resources.functions_propulsion import *
from tests.verification.model_verification.resources.functions_turbofan import *


class Turbofan:

    def __init__(self):
        self.h = None
        self.M = None
        self.mf = None
        self.bpr = None
        self.T0_4 = None
        self.dT0_4 = None
        self.T_amb = None
        self.p_amb = None

        # Intercooler and recuperator
        self.taic_core = None
        self.taic_bp = None
        self.rec_dt = None

        self.E_intercooler = None
        self.E_recuperator = None

        self.pi_intercooler = None
        self.pi_recuperator_comp = None
        self.pi_recuperator_nozz = None

        # pressure ratios
        self.pi_fan = None
        self.pi_LPC = None
        self.pi_HPC = None
        self.pi_cc = None
        self.pi_inlet = 1
        # self.pi_nozzle_core = None
        # self.pi_nozzle_bypass = None

        # isentropic efficiencies
        self.nu_inlet = 1
        self.nu_fan = None
        self.nu_LPC = None
        self.nu_HPC = None
        self.nu_LPT = None
        self.nu_HPT = None
        self.nu_cc = None
        self.nu_comb = None

        # nozzle efficiencies
        self.nu_nozzle = None
        self.nu_nozzle_core = None
        self.nu_nozzle_bypass = None

        # mechanical efficiencies
        self.nu_mech = None
        self.nu_mech_LP = None
        self.nu_mech_HP = None
        self.nu_gearbox = None

        # constants
        self.LHV = 43e6
        self.cp_air = 1000
        self.cp_gas = 1150
        self.k_air = 1.4
        self.k_gas = 1.33
        self.R = 287

        self.w_LPS = None

    def efficiencies(self):
        # nozzle efficiencies
        self.nu_nozzle_core = self.nu_nozzle_core if not isinstance(self.nu_nozzle_core, type(None)) else self.nu_nozzle
        self.nu_nozzle_bypass = self.nu_nozzle_bypass if not isinstance(self.nu_nozzle_bypass, type(None)) else self.nu_nozzle

        # combustion chamber efficiency
        if not isinstance(self.nu_comb, type(None)):
            self.nu_cc = self.nu_cc * self.nu_comb
        elif isinstance(self.nu_cc, type(None)):
            self.nu_cc = self.nu_comb

        # mechanical efficiencies
        if isinstance(self.nu_mech_LP, type(None)):
            self.nu_mech_LP = self.nu_mech

        if isinstance(self.nu_mech_HP, type(None)):
            self.nu_mech_HP = self.nu_mech

    def ambient(self):
        # initialize efficiencies
        self.efficiencies()
        # initialize p, t
        self.p, self.t = self.p_amb, self.T_amb
        self.record = [[self.p, self.t]]

        self.p0_0, self.t0_0 = ambient(self.p, self.t, self.M, self.k_air)

        self.method_record(self.p0_0, self.t0_0)

    def inlet(self):
        self.p0_2, self.t0_2 = inlet(self.p_amb, self.T_amb, self.M, self.k_air, self.pi_inlet, self.nu_inlet)

        self.method_record(self.p0_2, self.t0_2)

    def fan(self, stage):
        self.p0_21, self.t0_21 = fan(self.p, self.t, self.k_air, self.pi_fan, self.nu_fan,
                                     stage=stage)

        self.method_record(self.p0_21, self.t0_21)

    def hmf(self):
        self.hmf = hot_mf(self.bpr, self.mf)
        self.cmf = self.mf - self.hmf

    def lpc(self, stage):
        self.p0_22, self.t0_22 = compressor(pbc=self.p, tbc=self.t, k=self.k_air, pi=self.pi_LPC, nu=self.nu_LPC,
                                            stage=stage, kind='LPC')

        self.method_record(self.p0_22, self.t0_22)

    def intercooler(self, stage):
        self.f = intercooler_f(self.E_intercooler, self.hmf, self.cmf,
                               self.t, self.taic_core,
                               self.t0_21, self.taic_bp)
        self.p0_23, self.t0_23 = intercooler_t_given(self.p, self.taic_core, self.pi_intercooler, stage)

        self.method_record(self.p0_23, self.t0_23)

    def hpc(self, stage):
        self.p0_24, self.t0_24 = compressor(pbc=self.p, tbc=self.t, k=self.k_air, pi=self.pi_HPC, nu=self.nu_HPC,
                                            stage=stage, kind='HPC')

        self.method_record(self.p0_24, self.t0_24)

    def recuperator_heat(self, stage):
        self.p0_3, self.t0_3 = recuperator_heat(self.E_recuperator, self.t, self.rec_dt, self.cp_air, self.cp_gas,
                                                self.p, self.pi_recuperator_comp, stage)

        self.method_record(self.p0_3, self.t0_3)

    def cc(self, kind, stage):
        if kind == 't_given':
            self.p0_4, self.t0_4 = cc_t_given(pbcc=self.p, tacc=self.T0_4, pi=self.pi_cc,
                                              stage=stage)
        if kind == 'dt_given':
            self.p0_4, self.t0_4 = cc_dt_given(pbcc=self.p, tbcc=self.t, dt=self.dT0_4, pi=self.pi_cc,
                                               stage=stage)
        if kind == 'fmf_given':
            self.p0_4, self.t0_4 = cc_fmf_given(pbcc=self.p, tbcc=self.t, pi=self.pi_cc, mf=self.hmf, fmf=self.fmf,
                                                nu=self.nu_cc, cp=self.cp_gas, LHV=self.LHV, stage=stage)

        self.method_record(self.p0_4, self.t0_4)

    def fmf(self):
        self.fmf = fuel_mass_flow(tbc=self.t0_3, tac=self.t0_4, mf=self.hmf, cp=self.cp_gas, nu=self.nu_cc, lhv=self.LHV)

    def work_exerted(self):
        self.w_LPC = compr_work(self.hmf, self.cp_air, self.t0_21, self.t0_22, 'LPC')
        self.w_HPC = compr_work(self.hmf, self.cp_air, self.t0_23, self.t0_24, 'HPC')
        self.w_fan = fan_power(self.t0_2, self.t0_21, self.mf, self.cp_air, self.nu_mech_LP, None)

    def work_required(self):
        fan_w = self.w_fan/self.nu_gearbox if not isinstance(self.nu_gearbox, type(None)) else self.w_fan
        self.w_HPT = turbine_pr(self.w_HPC, self.nu_mech_HP, which='HPT')
        self.w_LPT = turbine_pr(self.w_LPC + fan_w, self.nu_mech_LP)

    def hpt(self, stage):
        self.p0_45, self.t0_45 = turbine(w=self.w_HPT, mf=self.hmf+self.fmf, cp=self.cp_gas, k=self.k_gas, nu=self.nu_HPT,
                                         tbt=self.t, pbt=self.p,
                                         stage=stage, kind='HPT')

        self.method_record(self.p0_45, self.t0_45)

    def lpt(self, stage):
        self.p0_5, self.t0_5 = turbine(w=self.w_LPT, mf=self.hmf+self.fmf, cp=self.cp_gas, k=self.k_gas, nu=self.nu_LPT,
                                       tbt=self.t, pbt=self.p,
                                       stage=stage, kind='LPT')

        self.method_record(self.p0_5, self.t0_5)

    def recuperator_remove(self, stage):
        self.p0_7, self.t0_7 = recuperator_remove(self.t, self.rec_dt,
                                                  self.p, self.pi_recuperator_nozz, stage)

        self.method_record(self.p0_7, self.t0_7)

    def pi_nozzle_core(self):
        self.pi_nozzle_core = pi_nozzle(self.p, self.p_amb, 'Core')

    def pi_nozzle_bypass(self):
        self.pi_nozzle_bypass = pi_nozzle(self.p0_21, self.p_amb, 'Bypass')

    def nozzle_core(self, stage):
        print_color('Core flow', 'green')

        pi_nozzle_core = self.pi_nozzle_core if not isinstance(self.pi_nozzle_core, type(None)) else self.p / self.p_amb

        p_pc, self.choked_core = is_nozzle_choked(pi_nozzle_core, self.nu_nozzle_core, self.k_gas)

        if self.choked_core is True:
            self.p8, self.t8 = choked_nozzle(self.p, self.t, p_pc, self.k_gas, self.nu_nozzle_core,
                                             stage=stage)
        else:
            self.p8, self.t8 = non_choked_nozzle(self.p, self.t, self.p_amb, self.k_gas, self.nu_nozzle_core,
                                                 stage=stage)

        self.method_record(self.p8, self.t8)

    def nozzle_bypass(self, stage):
        print_color('Bypass flow', 'green')

        pi_nozzle_bypass = self.pi_nozzle_bypass if not isinstance(self.pi_nozzle_bypass, type(None)) else self.p0_21 / self.p_amb

        p_pc, self.choked_bypass = is_nozzle_choked(pi_nozzle_bypass, self.nu_nozzle_bypass, self.k_air)

        if self.choked_bypass is True:
            self.p18, self.t18 = choked_nozzle(self.p0_21, self.t0_21, p_pc, self.k_air, self.nu_nozzle_bypass,
                                               stage=stage)
        else:
            self.p18, self.t18 = non_choked_nozzle(self.p0_21, self.t0_21, self.p_amb, self.k_air, self.nu_nozzle_bypass,
                                                   stage=stage)

    def exit_velocity(self):
        self.v0 = v0(self.k_air, self.R, self.T_amb, self.M)
        if self.choked_core is True:
            self.v8 = exit_velocity_choked(self.cp_gas, self.t0_5, self.t8, 'Core')
        else:
            self.v8 = exit_velocity_unchoked(self.k_gas, self.R, self.t8, 'Core')

        if self.choked_bypass is True:
            self.v18 = exit_velocity_choked(self.cp_air, self.t0_21, self.t18, 'Bypass')
        else:
            self.v8 = exit_velocity_unchoked(self.k_air, self.R, self.t18, 'Bypass')

    def nozzle_exit_area(self, stage):
        self.A8 = A_exit(self.hmf+self.fmf, self.t8, self.p8, self.v8, self.R, stage)
        self.A18 = A_exit(self.cmf, self.t18, self.p18, self.v18, self.R, '1{}'.format(stage))

    def thrust_core(self):
        self.T_core = flow_thrust(self.hmf+self.fmf, self.v8, self.v0, self.A8, self.p8, self.p_amb, self.choked)

    def thrust_fan(self):
        self.T_fan = flow_thrust(self.cmf, self.v18, self.v0, self.A18, self.p18, self.p_amb, self.choked, which='Fan')

    def sfc(self):
        self.sfc = sfc(self.fmf, self.T_core+self.T_fan)

    def method_record(self, p, t):
        self.p = p
        self.t = t
        self.record.append([self.p, self.t])

    def record_correction(self, i, p=None, t=None):
        if not isinstance(p, type(None)):
            self.record[i][0] = p
        if not isinstance(t, type(None)):
            self.record[i][1] = t



from copy import deepcopy

from huracan.constants import R
from huracan.engine import stream, component
from huracan.thermo.processes import absolute, diffusion, compression, combustion, expansion


class fluid:
    """
    Fluid
    -----
    """
    @classmethod
    def _diversion(cls, f, fr):
        """
        Fluid diversion.
        - Diverted flow mass flow:       fraction*mf
        - Main flow mass flow:       (1-fraction)*mf

        :param f:  Fluid to be diverted.
        :param fr: Fraction of diverted fluid.

        :type f:   mixture or fluid
        :type fr:  0 <= float <= 1

        :return:   [fluid] Main flow
                   [fluid] Diverted flow
        """
        mf, df = deepcopy(f), deepcopy(f)

        mf.mf *= 1 - fr
        df.mf *= fr

        return mf, df

    def __mul__(self, other):
        """
        Return two flows representing the diversion.
        """
        return self._diversion(self, other)

    def __rmul__(self, other):
        """
        Return two flows representing the diversion.
        """
        return self._diversion(self, other)


class gas(fluid):
    """
    Ideal gas model
    ---------------

    All gas phenomena are modelled using its
    - Mass flow
    - Constant pressure specific heat capacity
    - Ratio of specific heat capacities
    """
    def __init__(self, mf, cp, k,
                 m, t_0, p_0):
        """
        :param mf:  [kg] Mass flow
        :param cp:  [-]  Constant pressure specific heat
        :param k:   [-]  Specific heat ratio
        :param m:   [M]  Flow Mach number      |
        :param t_0: [K]  Initial temperature   | -> Total pressure and temperature calculated from inputs
        :param p_0: [Pa] Initial pressure      |

        :type mf:   float
        :type cp:   (T: float) -> float
        :type k:    (T: float) -> float
        :type m:    float
        :type t_0:  float
        :type p_0:  float
        """
        self.mf  = mf
        self.cp  = cp
        self.k   = k

        self.m   = m
        self.v_0 = m*(k(t_0)*R*t_0)**0.5

        self.t_0 = t_0
        self.p_0 = p_0

        self.absolute()

    def state(self):
        """
        Update gas state variables.
        - V - Specific volume
        - S - Specific entropy
        - H - Specific enthalpy
        """
        self.V = self.t0*R/self.p0
        self.S = 1                          #TODO: implement
        self.H = 1                          #TODO: implement

    def __add__(self, other):
        """
        Mixture creation operator: <gas> + <gas>

        In the case of adding fuel to a gas, the
        following assumptions are considered:
        - Total temperature and pressure of gas are
          unaltered.
        - Constant pressure specific heat capacity Cp
          of gas is unaltered.
        - Ratio of specific heats k of gas is unaltered.

        :type other: gas or fuel
        """
        if isinstance(other, gas):
            return mixture(self, other)
        if isinstance(other, fuel):
            self.mf += other.mf
            return self

    def __sub__(self, other):
        """
        Stream creation operator: <gas> + <component/stream>

        :type other: component or stream

        :return:     stream instance
        """
        if isinstance(other, component):
            s = stream(self)-other
            return s
        elif isinstance(other, stream):
            other.gas = self
            return other

    def absolute(self):
        """
        Returns the absolute temperature and pressure of a gas
        flowing at a certain Mach number.

        :return: process data class instance
        """

        k   = self.k(self.t_0)

        p = absolute(k   = k,
                     m   = self.m,
                     t_0 = self.t_0,
                     p_0 = self.p_0)

        self.t0 = p.t0
        self.p0 = p.p0

        self.state()

        return p

    def diffusion(self, eta,
                  PI=None,
                  TAU=None):
        """
        Diffusion
        ---------

        Assumptions:
        - Adiabatic
        - Reversible
        - Temperature remains constant through process

        The absolute pressure change is determined by the total pressure ratio
        _PI_ if provided. Else, _PI_ is calculated using the isentropic
        efficiency _nu_.

        :param eta: [-]  Isentropic efficiency
        :param PI:  [-]  Pressure ratio

        :type eta:  float
        :type PI:   float

        :return:    process instance
        """

        k   = self.k(self.t0)
        cp  = self.cp(self.t0)

        p = diffusion(mf  = self.mf,
                      cp  = cp,
                      k   = k,
                      m   = self.m,
                      t_0 = self.t_0,
                      p_0 = self.p_0,
                      eta = eta,
                      PI  = PI,
                      TAU = TAU,
                      t0  = self.t0 if hasattr(self, 't0') else None,
                      p0  = self.p0 if hasattr(self, 'p0') else None
                      )

        self.t0 = p.t01
        self.p0 = p.p01

        self.state()

        return p

    def compression(self, eta,
                    PI=None,
                    TAU=None,
                    ):
        """
        Compression
        -----------

        Assumptions:
        - Adiabatic
        - Reversible

        Input pairs:
            - eta, PI
                - TAU is calculated
            - eta, TAU
                - PI is calculated

        :param eta: [-]  Isentropic efficiency
        :param PI:  [-]  Pressure ratio
        :param TAU: [-]  Temperature ratio

        :type eta:  float
        :type PI:   float
        :type TAU:  float

        :return:    process instance
        """

        k   = self.k(self.t0)
        cp  = self.cp(self.t0)

        p   = compression(mf  = self.mf,
                          cp  = cp,
                          k   = k,
                          t00 = self.t0,
                          p00 = self.p0,
                          eta = eta,
                          PI  = PI,
                          TAU = TAU,
                          )

        self.t0 = p.t01
        self.p0 = p.p01

        self.state()

        return p

    def expansion(self, eta,
                  PI=None,
                  TAU=None,
                  ):
        """
        Expansion
        ---------

        Assumptions:
        - Adiabatic
        - Reversible

        Input pairs:
            - PI, eta
                - TAU is calculated
            - TAU, eta
                - PI is calculated

        :param eta: [-]  Isentropic efficiency
        :param PI:  [-]  Pressure ratio
        :param TAU: [-]  Temperature ratio

        :type eta:  float
        :type PI:   float
        :type TAU:  float

        :return:    process instance
        """

        k   = self.k(self.t0)
        cp  = self.cp(self.t0)

        p   = expansion(mf  = self.mf,
                        cp  = cp,
                        k   = k,
                        t00 = self.t0,
                        p00 = self.p0,
                        eta = eta,
                        PI  = PI,
                        TAU = TAU,
                        )

        self.t0 = p.t01
        self.p0 = p.p01

        self.state()

        return p

    def heat_addition(self,
                      eta,
                      cp,
                      fuel_mf,
                      fuel_LHV,
                      PI=1):
        """
        Constant pressure heat addition
        -------------------------------

        Assumptions:
        - Perfect heat addition
        - Constant pressure

        :param eta:      Isentropic efficiency
        :param cp:       Specific heat of gas during heat addition
        :param fuel_mf:  Fuel mass flow
        :param fuel_LHV: Fuel Lower Heating Value (heat of combustion)
        """

        p = combustion(mf       = self.mf,
                       cp       = cp,
                       t00      = self.t0,
                       p00      = self.p0,
                       fuel_mf  = fuel_mf,
                       fuel_LHV = fuel_LHV,
                       eta      = eta,
                       PI       = PI)

        self.t0 = p.t01
        self.p0 = p.p01

        self.state()

        return p


class fuel:
    """
    Fuel model
    ----------
    """
    def __init__(self, LHV,
                 mf=None):
        """
        :param LHV: Fuel Lower Heating Value (heat of combustion)
        :param mf:  Fuel mass flow
        """
        self.LHV = LHV
        self.mf = mf


class mixture(gas):
    """
    Mixture model
    -------------
    """

    @classmethod
    def _mix_t0(cls, gas1, gas2):
        """
        Total temperature of fluid mixture.
        """
        n = gas1.mf*gas1.cp(gas1.t0)*gas1.t0 + gas2.mf*gas2.cp(gas2.t0)*gas2.t0
        d = gas1.mf*gas1.cp(gas1.t0) + gas2.mf*gas2.cp(gas2.t0)
        t0_f = n / d
        return t0_f

    @classmethod
    def _mix_p0(cls, gas1, gas2):
        """
        Total pressure of fluid mixture.
        """
        n = gas1.p0*gas1.mf + gas2.p0*gas2.mf
        d = gas1.mf + gas2.mf
        p0_f = n / d
        return p0_f

    @classmethod
    def _mix_cp(cls, gas1, gas2):
        """
        Constant pressure specific heat at of fluid mixture.
        """
        cp_f = lambda t: (gas1.cp(t)*gas1.mf + gas2.cp(t)*gas2.mf) / (gas1.mf + gas2.mf)
        return cp_f

    @classmethod
    def _mix_k(cls, gas1, gas2):
        """
        Specific heat ratio of fluid mixture.
        """
        cp_f = lambda t: (gas1.k(t)*gas1.mf + gas2.k(t)*gas2.mf) / (gas1.mf + gas2.mf)
        return cp_f

    def __init__(self, gas1, gas2):
        """
        :param gas1: Mixture fluid 1
        :param gas2: Mixture fluid 2

        :type gas1:  gas or mixture
        :type gas2:  gas or mixture
        """

        mf = gas1.mf + gas2.mf
        super().__init__(mf  = mf,
                         cp  = self._mix_cp(gas1, gas2),
                         k   = self._mix_k(gas1, gas2),
                         m   = 0,
                         t_0 = self._mix_t0(gas1, gas2),
                         p_0 = self._mix_p0(gas1, gas2))



"""
Huracan
-------
Single-spool turboprop engine test.
"""

from huracan.engine import shaft
from huracan.thermo.fluids import gas, fuel
from huracan.components import inlet, prop, compressor, combustion_chamber, turbine, afterburner, nozzle, power_plant

from tests.utils import verify

mf = 35
M  = 0.5
t  = 288.15
p  = 101325.00

fuel = fuel(LHV=43e6)

g = gas(mf=mf,
        cp=lambda T: 1150 if T > 600 else 1000,
        k=lambda T: 1.33 if T > 600 else 1.4,
        m=M, t_0=t, p_0=p)

pr = prop(eta=0.9, w=3.91e6, PI=1)
i  = inlet(PI=1)
c  = compressor(eta=0.85, PI=11.5)
cc = combustion_chamber(fuel=fuel, eta=0.995, t01=1130, PI=0.96)
t  = turbine(eta=0.89)
n  = nozzle(eta=0.95)

shaft1 = shaft(pr, c, t, eta=0.99)

stream = g-pr-i-c-cc-t-n

stream.run()

"""
Verification

- Error must stay below 0.1% of the verification value 
"""
# Total temperature
verify(stream['0.il'].t0, 302.56)               # Inlet
verify(stream['0.cp'].t0, 661.84)               # Compressor

# Total pressure
verify(stream['0.il'].p0, 120192.99554985)      # Inlet
verify(stream['0.cp'].p0, 1382219.44882326)     # Compressor

"""
Exceptions:

The fuel mass flow method differs from
that used by the verification model. 
This causes differences in the models'
results after the main combustion chamber.

- Error must stay below 5% of the verification value 
"""
# Fuel mass flow
verify(cc.fuel.mf, 0.4425, 0.05)

# Total temperature
verify(stream['0.tb'].t0, 721.35,  0.05)        # Turbine
verify(stream['0.nz'].t0, 645.8)                # Nozzle

# Total pressure
verify(stream['0.tb'].p0, 162.37e3, 0.05)       # Turbine
verify(stream['0.nz'].p0, 101325)               # Nozzle

# Exit velocity
verify(stream.v_exit(), 417.2)




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="huracan",
    version="0.0.a2",
    author="Antonio Lopez Rivera",
    author_email="antonlopezr99@gmail.com",
    description="Open source, 0-dimensional, object-oriented airbreathing engine modelling package for preliminary analysis and design of airbreathing engines, divulgation and educational purposes.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/alopezrivera/huracan",
    packages=setuptools.find_packages(),
    install_requires=[
        "Python-Alexandria",
        "mpl_plotter"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)



from tests.verification.model_verification.resources.utils_propulsion import *


def hot_mf(bpr, mf):
    print_color('Hot mass flow', 'green')
    print(units('Core ṁ = {:,.4f}'.format(mf/(bpr+1)), 'Kg/s'))
    return mf/(bpr+1)


def fan(pbf, tbf, k, pi, nu, stage=None):
    """
    Total pressure and temperature through turbofan fan
    :param pbf: Total pressure ante fan
    :param tbf: Total temperature ante fan
    :param k: Specific heat capacity of gas flowing through compressor
    :param pi: Compressor pressure ratio
    :param nu: Compressor isentropic efficiency
    :param stage: Stage
    :return: Total pressure and temperature after fan stage
    """
    p = pi*pbf
    t = tbf*(1+1/nu*(pi**((k-1)/k)-1))
    print_color('Fan', 'green')
    print(units('T0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', t), 'K')
          +
          units('\np0_{} = {:,.4f}'.format(stage if not isinstance(stage, type(None)) else 'n+1', p), 'Pa')
          )
    return p, t


def fan_power(tbf, taf, mf, cp):
    """
    Worked done on flow by fan
    :param tbf: Total temperature before fan
    :param taf: Total temperature after fan
    :param mf: ṁ through fan
    :param cp: Specific heat capacity of flow through fan
    :return: Work required by the flow on turbine
    """
    print_color('Fan power required', 'yellow')
    print(units('Ẇ_fan = {:,.4f}'.format(mf*cp*(taf-tbf)), 'W'))
    return mf*cp*(taf-tbf)



import unittest

from huracan.thermo.fluids import gas, fuel

mf = 700
m  = 0.6
t  = 288
p  = 101325
fr = 0

g = gas(mf = mf,
        cp = lambda T: 1150 if T > 600 else 1000,
        k  = lambda T: 1.33 if T > 600 else 1.4,
        m  = m, t_0=t, p_0=p)

f = gas(mf = mf,
        cp = lambda T: 1150 if T > 600 else 1000,
        k  = lambda T: 1.33 if T > 600 else 1.4,
        m  = m, t_0=t*fr, p_0=p*fr)

k = gas(mf = mf,
        cp = lambda T: 1500 if T > 1000 else 1200 if T > 800 else 400,
        k  = lambda T: 1.33 if T > 600 else 1.4,
        m  = m, t_0=t*fr, p_0=p*fr)


class TestsFluid(unittest.TestCase):

    def test_mixture(self):
        global h
        h = g+f

        assert abs(h.t0 - g.t0/2) < 10e-12
        assert abs(h.p0 - g.p0/2) < 10e-12

    def test_diversion(self):
        mf = h.mf
        fr = 0.2
        i, j = fr*h

        assert i.mf == mf*(1-fr)
        assert j.mf == mf*fr
